import shutil

import pytest

import dclab
from dclab import RTDCWriter
from dclab.rtdc_dataset import rtdc_copy
from dclab.rtdc_dataset.fmt_http import RTDC_HTTP
import h5py
import numpy as np

from helper_methods import DCOR_AVAILABLE, retrieve_data


http_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
            "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
            "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")


def test_features_local_basic():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    with dclab.new_dataset(h5path) as ds:
        # access time and index so that they are in the local features
        assert ds["index"][0] == 1
        assert len(ds["time"]) == 10
        assert ds.features_local == ds.features_loaded


def test_features_local_basin():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("small.rtdc")

    with h5py.File(h5path, "a") as h5:
        del h5["events/deform"]

    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_copy(src_h5file=src,
                  dst_h5file=hw.h5file,
                  features=["area_um"])
        hw.store_basin(basin_name="example basin",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[h5path],
                       basin_descr="an example test basin",
                       )

    with dclab.new_dataset(h5path_small) as ds:
        # access time and index so that they are in the local features
        assert "area_um" in ds.features_local
        assert "area_msd" in ds.features_local
        # not accessed, thus not locally available
        assert "time" not in ds.features_local
        assert "time" in ds.features_loaded
        # area_um was removed above, but can be computed (ancillary)
        assert "deform" in ds.features_loaded
        assert "deform" not in ds.features_local
        assert "circ" in ds.features_local
        assert ds["deform"] is not None  # just access the feature
        assert "deform" in ds.features_local


@pytest.mark.skipif(not DCOR_AVAILABLE, reason="DCOR not accessible")
def test_features_local_remote():
    """Open a remote dataset and see whether local features are empty"""
    pytest.importorskip("requests")
    with RTDC_HTTP(http_url) as ds:
        assert not ds.features_local
        assert ds.features_loaded
        assert ds["deform"] is not None  # access a feature
        assert ds.features_local == ["deform"]


@pytest.mark.skipif(not DCOR_AVAILABLE, reason="DCOR not accessible")
def test_features_local_remote_basin(tmp_path):
    pytest.importorskip("requests")
    tmp_path = tmp_path.resolve()
    h5path = tmp_path / "test_basin_http.rtdc"

    with h5py.File(h5path, "a") as dst, RTDC_HTTP(http_url) as src:
        # Store non-existent basin information
        with RTDCWriter(dst, mode="append") as hw:
            meta = src.config.as_dict(pop_filtering=True)
            hw.store_metadata(meta)
            hw.store_basin(basin_name="example basin",
                           basin_type="remote",
                           basin_format="http",
                           basin_locs=[http_url],
                           basin_descr="an example http test basin",
                           )

    with dclab.new_dataset(h5path) as ds:
        assert not ds.features_local
        assert ds.features_loaded
        assert ds["deform"] is not None  # access a feature
        # The "circ" feature should not be downloaded for accessing the
        # "deform" feature.
        assert ds.features_local == ["deform"]
        assert "deform" not in ds.features_innate
        assert "deform" in ds.features_basin
        assert ds.basins[0].ds.features_local == ["deform"]


def test_numpy2_copy_if_needed():
    """Test numpy inconsistent copy keyword argument

    This test did not pass with numpy>2,<2.1.
    This test passed with numpy>=2.1 before introducing `copy_if_needed`
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path2 = h5path.with_name("test.rtdc")
    shutil.copy2(h5path, h5path2)

    ds1 = dclab.new_dataset(h5path)
    ds2 = dclab.new_dataset(h5path2)
    assert len(ds1) == len(ds2)

    for feat in ds1.features_scalar:
        # Accessing the __array__ method here failed because numpy 1 and 2
        # expected different values for the copy keyword argument.
        # https://numpy.org/devdocs/numpy_2_0_migration_guide.html#adapting-to-changes-in-the-copy-keyword.
        assert np.allclose(ds1[feat][:], ds2[feat][:],
                           atol=1e-6,
                           rtol=0)



"""Exemplary plugin feature

You can import the features defined in this file into dclab
with ``dclab.load_plugin_feature("/path/to/plugin_example.py")``.
"""


def compute_some_new_features(rtdc_ds):
    """The function that does the heavy-lifting"""
    circ_per_area = rtdc_ds["circ"] / rtdc_ds["area_um"]
    circ_times_area = rtdc_ds["circ"] * rtdc_ds["area_um"]
    # returns a dictionary-like object
    return {"circ_per_area": circ_per_area, "circ_times_area": circ_times_area}


info = {
    "method": compute_some_new_features,
    "description": "This plugin will compute some features",
    "long description": "Even longer description that "
                        "can span multiple lines",
    "feature names": ["circ_per_area", "circ_times_area"],
    "feature labels": ["Circularity per Area", "Circularity times Area"],
    "features required": ["circ", "area_um"],
    "config required": [],
    "method check required": lambda x: True,
    "scalar feature": [True, True],
    "version": "0.1.0",
}



import collections

import numpy as np

from dclab.rtdc_dataset.export import yield_filtered_array_stacks


def test_yield_filtered_array_stacks_array():
    enum = np.arange(547)
    ebol = np.ones(547, dtype=bool)
    ebol[10] = False
    ebol[412] = False

    data = np.random.random((547, 80, 320))
    indices = enum[ebol]
    stacked = []
    for chunk in yield_filtered_array_stacks(data, indices):
        stacked.append(chunk)
    assert len(stacked) == 55
    assert len(stacked[0]) == len(stacked[1])
    assert len(stacked[-1]) == 547 - 2 - 54 * len(stacked[0])

    data2 = np.concatenate(stacked, axis=0)
    assert len(data2) == 547 - 2
    assert data2[0].shape == data[0].shape
    assert np.all(data[indices] == data2)


def test_yield_filtered_array_stacks_list():
    # custom list we will be using, which implements shape and dtype, but
    # no __array__.
    class ListFeat(collections.UserList):
        def __init__(self, data):
            super(ListFeat, self).__init__(data)

        @property
        def shape(self):
            return tuple([len(self.data)] + list(self.data[0].shape))

        @property
        def dtype(self):
            return self.data[0].dtype

    enum = np.arange(547, dtype=int)
    ebol = np.ones(547, dtype=bool)
    ebol[10] = False
    ebol[412] = False

    # instead of the above case, create a list of arrays here.
    datalist = []
    for _ in range(547):
        datalist.append(np.random.random((80, 320)))
    data = ListFeat(datalist)
    assert data.shape == (547, 80, 320)
    assert data.dtype == datalist[0].dtype
    assert np.all(data[0] == datalist[0])

    indices = enum[ebol]
    stacked = []
    for chunk in yield_filtered_array_stacks(data, indices):
        stacked.append(np.array(chunk, copy=True))
    assert len(stacked) == 55
    assert len(stacked[0]) == len(stacked[1])
    assert len(stacked[-1]) == 547 - 2 - 54 * len(stacked[0])
    assert np.all(stacked[0][0] == datalist[0])

    data2 = np.concatenate(stacked, axis=0)
    assert len(data2) == 547 - 2
    assert data2[0].shape == data[0].shape
    assert np.all(np.array(datalist)[indices] == data2)



import numpy as np
import warnings

import pytest

from dclab.features import emodulus


def test_buyukurganci_cell_carrier():
    """Test values inspired from the table in Herold's paper"""
    ch_sizes = [15, 15, 15, 20, 20, 20, 20, 20, 30, 30, 30, 40, 40, 40]
    fl_rates = [0.016, 0.032, 0.048, 0.02, 0.04, 0.06,
                0.08, 0.12, 0.16, 0.24, 0.32, 0.32, 0.40, 0.60]
    temps = [24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24]
    eta_a = [
        5.4001970145102955,
        4.257647591503594,
        3.704914977001318,
        6.725414611510781,
        5.3024816031010324,
        4.6141074582608255,
        4.180606367836839,
        3.63787533192147,
        5.0023466326283845,
        4.352936348316222,
        3.94397260600829,
        5.3024816031010324,
        4.9118302388920805,
        4.274170894950141,
    ]
    eta_b = [
        6.384624807560792,
        4.799374333891269,
        4.061421656375954,
        8.309306638844017,
        6.246173301784982,
        5.285760549704508,
        4.695299212276897,
        3.9733491445425506,
        5.824144616254946,
        4.928623072237973,
        4.378056820977706,
        6.246173301784982,
        5.697847203911832,
        4.821745173172601,
    ]

    for c, f, t, a in zip(ch_sizes, fl_rates, temps, eta_a):
        eta = emodulus.viscosity.get_viscosity(medium="0.49% MC-PBS",
                                               channel_width=c,
                                               flow_rate=f,
                                               temperature=t,
                                               model='buyukurganci-2022')
        assert np.allclose(eta, a)

    for c, f, t, b in zip(ch_sizes, fl_rates, temps, eta_b):
        eta = emodulus.viscosity.get_viscosity(medium="0.59% MC-PBS",
                                               channel_width=c,
                                               flow_rate=f,
                                               temperature=t,
                                               model='buyukurganci-2022')
        assert np.allclose(eta, b)


def test_buyukurganci_cell_carrier_zero_flow_rate():
    eta = emodulus.viscosity.get_viscosity(medium="0.59% MC-PBS",
                                           channel_width=20,
                                           flow_rate=0,
                                           temperature=21.5,
                                           model='buyukurganci-2022')
    assert np.isnan(eta)


@pytest.mark.parametrize("medium", ["0.49% MC-PBS", "0.59% MC-PBS"])
def test_buyukurganci_mcpbs_range(medium):
    # test values
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(
            medium=medium, temperature=21, model="buyukurganci-2022")
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(
            medium=medium, temperature=38, model="buyukurganci-2022")
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(medium=medium,
                                         temperature=np.linspace(1, 10, 8),
                                         model="buyukurganci-2022")
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)


def test_herold_cell_carrier():
    """Test using table from Christophs script"""
    ch_sizes = [15, 15, 15, 20, 20, 20, 20, 20, 30, 30, 30, 40, 40, 40]
    fl_rates = [0.016, 0.032, 0.048, 0.02, 0.04, 0.06,
                0.08, 0.12, 0.16, 0.24, 0.32, 0.32, 0.40, 0.60]
    temps = [24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24]
    eta_a = [5.8, 4.6, 4.1, 7.1, 5.7, 5.0,
             4.5, 4.0, 5.4, 4.7, 4.3, 5.7, 5.3, 4.6]
    eta_b = [7.5, 5.8, 5.0, 9.4, 7.3, 6.3,
             5.7, 4.9, 6.9, 5.9, 5.3, 7.3, 6.7, 5.8]

    for c, f, t, a in zip(ch_sizes, fl_rates, temps, eta_a):
        eta = emodulus.viscosity.get_viscosity(medium="CellCarrier",
                                               channel_width=c,
                                               flow_rate=f,
                                               temperature=t,
                                               model="herold-2017")
        assert np.allclose(np.round(eta, 1), a)

    for c, f, t, b in zip(ch_sizes, fl_rates, temps, eta_b):
        eta = emodulus.viscosity.get_viscosity(medium="CellCarrier B",
                                               channel_width=c,
                                               flow_rate=f,
                                               temperature=t,
                                               model="herold-2017")
        assert np.allclose(np.round(eta, 1), b)


@pytest.mark.parametrize("medium", ["0.49% MC-PBS", "0.59% MC-PBS"])
def test_herold_cellcarrier_range(medium):
    # test values
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(
            medium=medium, temperature=15, model="herold-2017")
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(
            medium=medium, temperature=28, model="herold-2017")
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(medium=medium,
                                         temperature=np.linspace(1, 10, 8),
                                         model="herold-2017")
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)


def test_kestin_water():
    """Test with data from Kestin et al., J. Phys. Chem. 7(3) 1978"""
    ref = np.array([[0, 1791.5],
                    [5, 1519.3],
                    [10, 1307.0],
                    [15, 1138.3],
                    [20, 1002.0],
                    [25, 890.2],
                    [30, 797.3],
                    [35, 719.1],
                    [40, 652.7],
                    ])
    ref[:, 1] *= 1e-3  # uPas to mPas

    res = emodulus.viscosity.get_viscosity(medium="water",
                                           temperature=ref[:, 0])
    assert np.allclose(res, ref[:, 1], rtol=8e-5, atol=0)

    # Since water viscosity is not dependent on flow rate, setting
    # flow rate to zero should still work.
    res2 = emodulus.viscosity.get_viscosity(medium="water",
                                            temperature=ref[:, 0],
                                            flow_rate=0,
                                            )
    assert np.allclose(res2, ref[:, 1], rtol=8e-5, atol=0)


def test_kestin_water_range():
    # test values
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(medium="water", temperature=-1)
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(medium="water", temperature=41)
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(medium="water",
                                         temperature=np.arange(-2, 10))
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)



"""A wrapper around R with the lme4 package"""
from . import rsetup, wrapr  # noqa: F401
from .wrapr import Rlme4, bootstrapped_median_distributions  # noqa: F401
from .rsetup import (  # noqa: F401
    set_r_lib_path, get_r_path, get_r_version, require_lme4, set_r_path)



import uuid

import pytest


from dclab.http_utils import (
    is_http_url, is_url_available, HTTPFile)

from helper_methods import DCOR_AVAILABLE

pytest.importorskip("requests")

if not DCOR_AVAILABLE:
    pytest.skip("No connection to DCOR", allow_module_level=True)


def test_http_file_basic():
    """This is a basic seek and read test

    It checks whether seeking to a position, reading things and then
    seeking to the same position yields the same result. Since HTTPFile
    caches things, we create two separate HTTPFiles.
    """
    s3_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
              "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
              "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")
    f1 = HTTPFile(s3_url)
    f2 = HTTPFile(s3_url)

    f1.seek(5000000)
    d1 = f1.read(200)

    f2.read(200)
    f2.seek(5000050)
    f2.read(200)
    f2.seek(5000000)
    d2 = f2.read(200)
    assert d1 == d2


def test_http_file_does_not_exist():
    bad_s3_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
                  "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
                  "resource/fb7/19f/ef-ffff-ffff-ffff-f4002af916f0")
    f1 = HTTPFile(bad_s3_url)
    with pytest.raises(ValueError,
                       match="Server replied with status code 403 Forbidden"):
        f1._parse_header()


@pytest.mark.parametrize("url, avail", [
    ("https://objectstore.hpccloud.mpcdf.mpg.de/"
     "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
     "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0", (True, "none")),
    # "noexisting"
    ("https://objectstore.hpccloud.mpcdf.mpg.de/"
     "noexisting-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
     "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0", (False, "not found")),
    # invalid URL
    ("https://example.com", (False, "invalid")),
    # nonexistent host
    (f"http://{uuid.uuid4()}.com/bucket/resource", (False, "no connection")),
    (f"https://{uuid.uuid4()}.com/bucket/resource", (False, "no connection")),
])
def test_object_available(url, avail):
    act = is_url_available(url, ret_reason=True)
    assert act == avail


@pytest.mark.parametrize("url", [
    "ftp://example.com/bucket/key",  # wrong scheme
    "example.com/bucket/key",  # missing scheme
    "example.com:80",  # missing key
    ])
def test_regexp_http_url_invalid(url):
    assert not is_http_url(url)


@pytest.mark.parametrize("url", [
    "https://example.com/bucket/key",
    "https://example.com/bucket/key2/key3",
    "https://example.com:80/bucket/key",
    "https://example.com:443/bucket/key",
    "http://example.com:80/bucket/key",
    "http://example.com/bucket/key",
    "https://example.com/bucket",
    "https://example.com/bucket/",
])
def test_regexp_http_url_valid(url):
    assert is_http_url(url)



from __future__ import annotations

import atexit
from contextlib import ExitStack
import copy
import functools
import json
import pathlib
# replace this import when dropping support for Python 3.8
# from importlib import resources as importlib_resources
import importlib_resources
import warnings

import numpy as np

from ... import definitions as dfn


_file_manager = ExitStack()
atexit.register(_file_manager.close)

#: Dictionary of look-up tables that the user added via :func:`register_lut`.
EXTERNAL_LUTS = {}


@functools.lru_cache()
def get_internal_lut_names_dict():
    """Return list of internal lut names"""
    lutfiles = {}
    for pp in importlib_resources.files('dclab.features.emodulus').iterdir():
        name = pp.name
        if name.startswith("lut_") and name.endswith(".txt"):
            lutfiles[name[4:-4]] = name
    return lutfiles


def get_lut_path(path_or_id):
    """Find the path to a LUT

    path_or_id: str or pathlib.Path
        Identifier of a LUT. This can be either an existing path
        (checked first), or an internal identifier (see
        :func:`get_internal_lut_names_dict`).
    """
    internal_dict = get_internal_lut_names_dict()
    if path_or_id == "FEM-2Daxis":
        # backwards compatibility
        warnings.warn("'FEM-2Daxis' is deprecated. Please use 'LE-2D-FEM-19'!",
                      DeprecationWarning)
        path_or_id = "LE-2D-FEM-19"
    if pathlib.Path(path_or_id).exists():
        lut_path = pathlib.Path(path_or_id)
    elif path_or_id in internal_dict:
        ref = importlib_resources.files(
            "dclab.features.emodulus") / internal_dict[path_or_id]
        lut_path = _file_manager.enter_context(
            importlib_resources.as_file(ref))
    elif path_or_id in EXTERNAL_LUTS:
        lut_path = EXTERNAL_LUTS[path_or_id]
    else:
        raise ValueError("File or LUT identifier does not exist: "
                         + "'{}'".format(path_or_id))
    return lut_path


def load_lut(lut_data: str | pathlib.Path | np.ndarray = "LE-2D-FEM-19"):
    """Load LUT data from disk

    Parameters
    ----------
    lut_data: path, str, or tuple of (np.ndarray of shape (N, 3), dict)
        The LUT data to use. If it is in :func:`get_internal_lut_names_dict()`,
        then the respective LUT will be used. Otherwise, a path to a
        file on disk or a tuple (LUT array, metadata) is possible.

    Returns
    -------
    lut: np.ndarray of shape (N, 3)
        The LUT data for interpolation
    meta: dict
        The LUT metadata

    Notes
    -----
    If lut_data is a tuple of (lut, meta), then nothing is actually
    done (this is implemented for user convenience).
    """
    if isinstance(lut_data, tuple):
        lut, meta = lut_data
        lut = np.array(lut, copy=True)  # copy, because of normalization
        meta = copy.deepcopy(meta)  # copy, for the sake of consistency
    elif isinstance(lut_data, (str, pathlib.Path)):
        lut_path = get_lut_path(lut_data)
        lut, meta = load_mtext(lut_path)
    else:
        raise ValueError("`name_path_arr` must be path, identifier, or tuple "
                         f"of (array, dict), got '{lut_data}'!")
    return lut, meta


def load_mtext(path):
    """Load column-based data from text files with metadata

    This file format is used for isoelasticity lines and look-up
    table data in dclab.

    The text file is loaded with `numpy.loadtxt`. The metadata
    are stored as a json string between the "BEGIN METADATA" and
    the "END METADATA" tags. The last comment (#) line before the
    actual data defines the features with units in square
    brackets and tab-separated. For instance:

        # [...]
        #
        # BEGIN METADATA
        # {
        #   "authors": "A. Mietke, C. Herold, J. Guck",
        #   "channel_width": 20.0,
        #   "channel_width_unit": "um",
        #   "date": "2018-01-30",
        #   "dimensionality": "2Daxis",
        #   "flow_rate": 0.04,
        #   "flow_rate_unit": "uL/s",
        #   "fluid_viscosity": 15.0,
        #   "fluid_viscosity_unit": "mPa s",
        #   "identifier": "LE-2D-ana-18",
        #   "method": "analytical",
        #   "model": "linear elastic",
        #   "publication": "https://doi.org/10.1016/j.bpj.2015.09.006",
        #   "software": "custom Matlab code",
        #   "summary": "2D-axis-symmetric analytical solution"
        # }
        # END METADATA
        #
        # [...]
        #
        # area_um [um^2]    deform    emodulus [kPa]
        3.75331e+00    5.14496e-03    9.30000e-01
        4.90368e+00    6.72683e-03    9.30000e-01
        6.05279e+00    8.30946e-03    9.30000e-01
        7.20064e+00    9.89298e-03    9.30000e-01
        [...]
    """
    path = pathlib.Path(path).resolve()

    # Parse metadata
    size = path.stat().st_size
    dump = []
    injson = False
    prev_line = ""
    with path.open("r", errors='replace') as fd:
        while True:
            line = fd.readline()
            if fd.tell() == size:
                # something went wrong
                raise ValueError("EOF: Could not parse '{}'!".format(path))
            elif len(line.strip()) == 0:
                # ignore empty lines
                continue
            elif not line.strip().startswith("#"):
                # we are done here
                if prev_line == "":
                    raise ValueError("No column header in '{}'!".format(
                        path))
                break
            elif line.startswith("# BEGIN METADATA"):
                injson = True
                continue
            elif line.startswith("# END METADATA"):
                injson = False
            if injson:
                dump.append(line.strip("#").strip())
            else:
                # remember last line for header
                prev_line = line
    # metadata
    if dump:
        meta = json.loads("\n".join(dump))
    else:
        raise ValueError("No metadata json dump in '{}'!".format(path))
    # header
    feats = []
    units = []
    for hh in prev_line.strip("# ").split("\t"):
        if hh.count(" "):
            ft, un = hh.strip().split(" ")
            un = un.strip("[]")
        else:
            ft = hh
            un = ""
        if not dfn.scalar_feature_exists(ft):
            raise ValueError("Scalar feature not known: '{}'".format(ft))
        feats.append(ft)
        units.append(un)
    # data
    data = np.loadtxt(path)

    meta["column features"] = feats
    meta["column units"] = units

    # sanity checks
    assert meta["channel_width_unit"] == "um"
    assert meta["flow_rate_unit"] == "uL/s"
    assert meta["fluid_viscosity_unit"] == "mPa s"
    for ft, un in zip(feats, units):
        if ft == "deform":
            assert un == ""
        elif ft == "area_um":
            assert un == "um^2"
        elif ft == "emodulus":
            assert un == "kPa"
        elif ft == "volume":
            assert un == "um^3"
        else:
            assert False, "Please add sanity check for {}!".format(ft)

    return data, meta


def register_lut(path, identifier=None):
    """Register an external LUT file in dclab

    This will add it to :const:`EXTERNAL_LUTS`, which is required
    for emodulus computation as an ancillary feature.

    Parameters
    ----------
    path: str or pathlib.Path
        Path to the external LUT file
    identifier: str or None
        The identifier is used for ancillary emodulus computation
        via the [calculation]: "emodulus lut" key. It is also used as
        the key in :const:`EXTERNAL_LUTS` during registration. If not
        specified, (default) then the identifier given as JSON metadata
        in `path` is used.
    """
    if identifier is None:
        _, md = load_mtext(path)
        try:
            identifier = md["identifier"]
        except KeyError:
            raise ValueError("The given LUT file '{}' does ".format(path)
                             + "not contain the 'identifier' keyword. You may "
                             + "specify it via the `identifier` keyword to "
                             + "this function.")
    if identifier in EXTERNAL_LUTS:
        raise ValueError("A LUT with an identifier '{}' ".format(identifier)
                         + "has already been registered!")
    elif identifier in get_internal_lut_names_dict():
        raise ValueError("The identifier '{}' is already ".format(identifier)
                         + "in use by an internal LUT!")
    EXTERNAL_LUTS[identifier] = path



"""Test S3 format"""
import time
import uuid

import numpy as np
import pytest


from dclab.rtdc_dataset.fmt_s3 import (
    is_s3_url, is_s3_object_available, RTDC_S3)

from helper_methods import DCOR_AVAILABLE


pytest.importorskip("boto3")


if not DCOR_AVAILABLE:
    pytest.skip("No connection to DCOR", allow_module_level=True)


def test_cache_features():
    # This is the calibration beads measurement.
    # https://dcor.mpl.mpg.de/dataset/figshare-7771184-v2/
    # resource/fb719fb2-bd9f-817a-7d70-f4002af916f0
    s3_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
              "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
              "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")

    with RTDC_S3(s3_url) as ds:
        t0 = time.perf_counter()
        _ = ds["deform"][:]
        t1 = time.perf_counter()
        for ii in range(10):
            _ = ds["deform"][:]
        t2 = time.perf_counter()
        assert t2 - t1 < t1 - t0


@pytest.mark.parametrize("url, avail", [
    ("https://objectstore.hpccloud.mpcdf.mpg.de/"
     "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
     "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0", True),
    # "noexisting"
    ("https://objectstore.hpccloud.mpcdf.mpg.de/"
     "noexisting-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
     "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0", False),
    # invalid URL
    ("https://example.com", False),
    # nonexistent host
    (f"http://{uuid.uuid4()}.com/bucket/resource", False),
    (f"https://{uuid.uuid4()}.com/bucket/resource", False),
])
def test_object_available(url, avail):
    act = is_s3_object_available(url)
    assert act == avail


def test_open_public_s3_dataset():
    # This is the calibration beads measurement.
    # https://dcor.mpl.mpg.de/dataset/figshare-7771184-v2/
    # resource/fb719fb2-bd9f-817a-7d70-f4002af916f0
    s3_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
              "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
              "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")

    with RTDC_S3(s3_url) as ds:
        assert ds.config["experiment"]["sample"] == "calibration_beads"
        assert len(ds) == 5000
        assert np.allclose(ds["deform"][100], 0.013640802,
                           atol=0,
                           rtol=1e-5)


@pytest.mark.parametrize("url", [
    "ftp://example.com/bucket/key",  # wrong scheme
    "example.com/bucket/key",  # missing scheme
    "https://example.com/bucket",  # missing key
    "https://example.com/bucket/",  # missing key
    "example.com:80",  # missing key and bucket
    ])
def test_regexp_s3_url_invalid(url):
    assert not is_s3_url(url)


@pytest.mark.parametrize("url", [
    "https://example.com/bucket/key",
    "https://example.com/bucket/key2/key3",
    "https://example.com:80/bucket/key",
    "https://example.com:443/bucket/key",
    "http://example.com:80/bucket/key",
    "http://example.com/bucket/key",
    ])
def test_regexp_s3_url_valid(url):
    assert is_s3_url(url)



"""Convert .tdms to .rtdc files"""
import argparse
import pathlib
import warnings

import hdf5plugin

from ..rtdc_dataset import fmt_tdms, new_dataset, RTDCWriter
from .._version import version

from . import common


def tdms2rtdc(path_tdms=None, path_rtdc=None, compute_features=False,
              skip_initial_empty_image=True, skip_final_empty_image=True,
              verbose=False):
    """Convert .tdms datasets to the hdf5-based .rtdc file format

    Parameters
    ----------
    path_tdms: str or pathlib.Path
        Path to input .tdms file
    path_rtdc: str or pathlib.Path
        Path to output .rtdc file
    compute_features: bool
        If `True`, compute all ancillary features and store them in the
        output file
    skip_initial_empty_image: bool
        In old versions of Shape-In, the first image was sometimes
        not stored in the resulting .avi file. In dclab, such images
        are represented as zero-valued images. If `True` (default),
        this first image is not included in the resulting .rtdc file.
    skip_final_empty_image: bool
        In other versions of Shape-In, the final image is sometimes
        also not stored in the .avi file. If `True` (default), this
        final image is not included in the resulting .rtdc file.
    verbose: bool
        If `True`, print messages to stdout
    """
    cmp_kw = hdf5plugin.Zstd(clevel=5)
    if path_tdms is None or path_rtdc is None:
        parser = tdms2rtdc_parser()
        args = parser.parse_args()

        path_tdms = pathlib.Path(args.tdms_path).resolve()
        path_rtdc = pathlib.Path(args.rtdc_path)
        compute_features = args.compute_features
        skip_initial_empty_image = not args.include_empty_boundary_images
        skip_final_empty_image = not args.include_empty_boundary_images
        verbose = True

    # Determine whether input path is a tdms file or a directory
    if path_tdms.is_dir():
        # we have a directory to search
        files_tdms = fmt_tdms.get_tdms_files(path_tdms)
        if path_rtdc.is_file():
            raise ValueError(
                f"Output path is a file, expected folder: '{path_rtdc}'!")
        files_rtdc = []
        for path_in in files_tdms:
            path_in = pathlib.Path(path_in)
            rp = path_in.relative_to(path_tdms)
            # determine output file name (same relative path)
            rpr = path_rtdc / rp.with_suffix(".rtdc")
            files_rtdc.append(rpr)
    else:
        # we have a single file or a non-existent path
        files_tdms = [path_tdms]
        files_rtdc = [path_rtdc]

    files_tdms, files_rtdc, files_temp = common.setup_task_paths(
        paths_in=files_tdms,
        paths_out=files_rtdc,
        allowed_input_suffixes=[".tdms"]
    )

    for ii in range(len(files_tdms)):
        path_in = files_tdms[ii]
        path_out = files_rtdc[ii]
        path_temp = files_temp[ii]

        if verbose:
            common.print_info(
                f"Converting {ii+1:d}/{len(files_tdms):d}: {path_in}")
        # create directory
        path_out.parent.mkdir(parents=True, exist_ok=True)
        # load and export dataset
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # ignore ResourceWarning: unclosed file <_io.BufferedReader...>
            warnings.simplefilter("ignore", ResourceWarning)  # noqa: F821
            # ignore SlowVideoWarning
            warnings.simplefilter("ignore",
                                  fmt_tdms.event_image.SlowVideoWarning)
            if skip_initial_empty_image:
                # If the initial frame is skipped when empty,
                # suppress any related warning messages.
                warnings.simplefilter(
                    "ignore",
                    fmt_tdms.event_image.InitialFrameMissingWarning)

            with new_dataset(path_in) as ds:
                # determine features to export
                if compute_features:
                    features = ds.features
                else:
                    # consider special case for "image", "trace", and "contour"
                    # (This will export both "mask" and "contour".
                    # The "mask" is computed from "contour" and it is needed
                    # by dclab for other ancillary features. We still keep
                    # "contour" because it is original data.
                    features = ds.features_innate

                common.skip_empty_image_events(
                    ds=ds,
                    initial=skip_initial_empty_image,
                    final=skip_final_empty_image)
                # export as hdf5
                ds.export.hdf5(path=path_temp,
                               features=features,
                               filtered=True,
                               override=True,
                               compression_kwargs=cmp_kw)

                # write logs
                custom_dict = {}
                # computed features
                cfeats = list(set(features) - set(ds.features_innate))
                if "mask" in features:
                    # Mask is always computed from contour data
                    cfeats.append("mask")
                custom_dict["ancillary features"] = sorted(cfeats)

                # command log
                logs = {"dclab-tdms2rtdc": common.get_command_log(
                    paths=[path_in], custom_dict=custom_dict)}
                # warnings log
                if w:
                    logs["dclab-tdms2rtdc-warnings"] = \
                        common.assemble_warnings(w)
                logs.update(ds.logs)
                with RTDCWriter(path_temp, compression_kwargs=cmp_kw) as hw:
                    for name in logs:
                        hw.store_log(name, logs[name])

                # Finally, rename temp to out
                path_temp.rename(path_out)


def tdms2rtdc_parser():
    descr = "Convert RT-DC .tdms files to the hdf5-based .rtdc file format. " \
            + "Note: Do not delete original .tdms files after conversion. " \
            + "The conversion might be incomplete."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('--compute-ancillary-features',
                        dest='compute_features',
                        action='store_true',
                        help='Compute features, such as volume or emodulus, '
                             + 'that are otherwise computed on-the-fly. '
                             + 'Use this if you want to minimize analysis '
                             + 'time in e.g. Shape-Out. CAUTION: ancillary '
                             + 'feature recipes might be subject to change '
                             + '(e.g. if an error is found in the recipe). '
                             + 'Disabling this option maximizes '
                             + 'compatibility with future versions and '
                             + 'allows to isolate the original data.')
    parser.set_defaults(compute_features=False)
    parser.add_argument('--include-empty-boundary-images',
                        dest='include_empty_boundary_images',
                        action='store_true',
                        help='In old versions of Shape-In, the first or last '
                             + 'images were sometimes not stored in the '
                             + 'resulting .avi file. In dclab, such images '
                             + 'are represented as zero-valued images. Set '
                             + 'this option, if you wish to include these '
                             + 'events with empty image data.')
    parser.set_defaults(include_empty_boundary_images=False)
    parser.add_argument('tdms_path', metavar="TDMS_PATH", type=str,
                        help='Input path (tdms file or folder containing '
                             + 'tdms files)')
    parser.add_argument('rtdc_path', metavar="RTDC_PATH", type=str,
                        help='Output path (file or folder), existing data '
                             + 'will be overridden')
    parser.add_argument('--version', action='version',
                        version=f'dclab-tdms2rtdc {version}')
    return parser



"""RT-DC dataset core classes and methods"""
import abc
import hashlib
import json
import os.path
import pathlib
import traceback
from typing import Literal
import uuid
import random
import warnings

import numpy as np

from .. import definitions as dfn
from .. import downsampling
from ..polygon_filter import PolygonFilter
from .. import kde_methods
from ..util import hashobj

from .feat_anc_core import AncillaryFeature, FEATURES_RAPID
from . import feat_basin
from .export import Export
from .filter import Filter


class FeatureShouldExistButNotFoundWarning(UserWarning):
    pass


class LogTransformWarning(UserWarning):
    pass


class RTDCBase(abc.ABC):
    def __init__(self, identifier=None, enable_basins=True):
        """RT-DC measurement base class

        Notes
        -----
        Besides the filter arrays for each data feature, there is a manual
        boolean filter array ``RTDCBase.filter.manual`` that can be edited
        by the user - a boolean value of ``False`` means that the event is
        excluded from all computations.
        """
        #: Local basins are basins that are defined on the user's file system.
        #: For reasons of data security (leaking data from a server or from a
        #: user's file system), dclab only allows remote basins (see
        #: :func:`basins_retrieve`) by default. This variable is set to True
        #: for the RTDC_HDF5 file format, because it implies the data are
        #: located on the user's computer.
        self._local_basins_allowed = False

        #: Dataset format (derived from class name)
        self.format = self.__class__.__name__.split("_")[-1].lower()

        # Cache attribute used for __len__()-function
        self._length = None
        self._polygon_filter_ids = []
        # Events have the feature name as keys and contain nD ndarrays.
        self._events = {}
        # Ancillaries have the feature name as keys and a
        # tuple containing feature and hash as value.
        self._ancillaries = {}
        # Temporary features are defined by the user ad hoc at runtime.
        self._usertemp = {}
        # List of :class:`.Basin` for external features
        self._basins = None
        # List of basin identifiers that should be ignored, used to
        # avoid circular basin dependencies
        self._basins_ignored = []
        # List of all features available via basins
        self._basins_features = None
        #: Configuration of the measurement
        self.config = None
        #: Export functionalities; instance of
        #: :class:`dclab.rtdc_dataset.export.Export`.
        self.export = Export(self)
        # Filtering functionalities; instance of
        # :class:`dclab.rtdc_dataset.filter.Filter`.
        self._ds_filter = None
        #: Dictionary of log files. Each log file is a list of strings
        #: (one string per line).
        self.logs = {}
        #: Dictionary of tables. Each table is an indexable compound numpy
        #: array.
        self.tables = {}
        #: Title of the measurement
        self.title = None
        #: Path or DCOR identifier of the dataset (set to "none"
        #: for :class:`RTDC_Dict`)
        self.path = None
        # Unique, random identifier
        if identifier is None:
            # Generate a unique, random identifier for this dataset
            rhex = [random.choice('0123456789abcdef') for _n in range(7)]
            self._identifier = "mm-{}_{}".format(self.format, "".join(rhex))
        else:
            self._identifier = identifier

        # Basins are initialized in the "basins" property function
        self._enable_basins = enable_basins

    def __contains__(self, feat):
        ct = False
        if (feat in self._events
                or feat in self._usertemp
                or feat in self.features_basin):
            ct = True
        else:
            # Check ancillary features data
            if feat in self._ancillaries:
                # already computed
                ct = True
            elif feat in AncillaryFeature.feature_names:
                # get all instance of AncillaryFeature that
                # check availability of the feature `feat`
                instlist = AncillaryFeature.get_instances(feat)
                for inst in instlist:
                    if inst.is_available(self):
                        # to be computed
                        ct = True
                        break
        return ct

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()

    def __getitem__(self, feat):
        if feat in self._events:
            return self._events[feat]
        elif feat in self._usertemp:
            return self._usertemp[feat]
        # 1. Check for cached ancillary data
        data = self._get_ancillary_feature_data(feat, no_compute=True)
        if data is not None:
            return data
        # 2. Check for h5dataset-based, file-based, or other basin data,
        #    in that order.
        for basin_type in ["internal", "file", None]:
            data = self._get_basin_feature_data(feat, basin_type=basin_type)
            if data is not None:
                return data
        # 3. Check for ancillary features that can be computed
        data = self._get_ancillary_feature_data(feat)
        if data is not None:
            return data
        if feat in self:
            warnings.warn(f"The feature {feat} is supposedly defined in "
                          f"{self}, but I cannot get its data. Please "
                          f"make sure you have not defined any unreachable "
                          f"remote basins.",
                          FeatureShouldExistButNotFoundWarning)
        # Not here ¯\_(ツ)_/¯
        raise KeyError(f"Feature '{feat}' does not exist in {self}!")

    def __iter__(self):
        """An iterator over all valid scalar features"""
        mycols = []
        for col in self._feature_candidates:
            if col in self:
                mycols.append(col)
        mycols.sort()
        for col in mycols:
            yield col

    def __len__(self):
        if self._length is None:
            self._length = self._get_length()
        return self._length

    def _get_length(self):
        # Try to get length from metadata.
        length = self.config["experiment"].get("event count")
        if length is not None:
            return length
        # Try to get the length from the feature sizes
        keys = list(self._events.keys()) or self.features_basin
        keys.sort()
        for kk in keys:
            length = len(self[kk])
            if length:
                return length
        else:
            raise ValueError(f"Could not determine size of dataset '{self}'.")

    def __repr__(self):
        repre = "<{} '{}' at {}".format(self.__class__.__name__,
                                        self.identifier,
                                        hex(id(self)))
        if self.path != "none":
            repre += " ({})>".format(self.path)
        return repre

    @property
    def basins(self):
        """Basins containing upstream features from other datasets"""
        if self._basins is None:
            if self._enable_basins:
                self._basins = self.basins_retrieve()
            else:
                self._basins = []
        return self._basins

    @property
    def filter(self):
        """Filtering functionalities; instance of :class:`.Filter`"""
        self._assert_filter()
        return self._ds_filter

    def _assert_filter(self):
        if self._ds_filter is None:
            self._ds_filter = Filter(self)

    def _get_ancillary_feature_data(self,
                                    feat: str,
                                    no_compute: bool = False):
        """Return feature data of ancillary features

        Parameters
        ----------
        feat: str
            Name of the feature
        no_compute: bool
            Whether to bother computing the feature. If it is not
            already computed, return None instead

        Returns
        -------
        data:
            The feature object (array-like) or None if it could not
            be found or was not computed.
        """
        data = None
        anhash = None
        if feat in AncillaryFeature.feature_names:
            # Try to find the feature in the ancillary features
            # (see feat_anc_core submodule for more information).
            # These features are cached in `self._ancillaries`.
            ancol = AncillaryFeature.available_features(self)
            if feat in ancol:
                # The feature is generally available.
                if feat in self._ancillaries:
                    # We have already computed the feature. Make sure that we
                    # have the updated one by checking the hash.
                    anhash = ancol[feat].hash(self)
                    if self._ancillaries[feat][0] == anhash:
                        # Use cached value
                        data = self._ancillaries[feat][1]
                # We either already have the ancillary feature or have to
                # compute it. We only compute it if we are asked to.
                if data is None and not no_compute:
                    anhash = anhash or ancol[feat].hash(self)
                    # Compute new value
                    data_dict = ancol[feat].compute(self)
                    for okey in data_dict:
                        # Store computed value in `self._ancillaries`.
                        self._ancillaries[okey] = (anhash, data_dict[okey])
                    data = data_dict[feat]
        return data

    def _get_basin_feature_data(
            self,
            feat: str,
            basin_type: Literal["file", "internal", "remote", None] = None):
        """Return feature data from basins

        Parameters
        ----------
        feat: str
            Name of the feature
        basin_type: str or bool
            The basin type to look at, which is either "file"-based
            (e.g. local on disk), "remote"-based (e.g. S3), or
            "internal"-type (e.g. h5py.Dataset inside the current HDF5 file).
            Defaults to `None` which means no preference.

        Returns
        -------
        data:
            The feature object (array-like) or None if it could not
            be found or was not computed.
        """
        data = None
        if self.basins:
            for bn in list(self.basins):
                if basin_type is not None and basin_type != bn.basin_type:
                    # User asked for specific basin type
                    continue
                try:
                    # There are all kinds of errors that may happen here.
                    # Note that `bn.features` can already trigger an
                    # availability check that may raise a ValueError.
                    # TODO:
                    #  Introduce some kind of callback so the user knows
                    #  why the data are not available. The current solution
                    #  (fail silently) is not sufficiently transparent,
                    #  especially when considering networking issues.
                    if feat in bn.features:
                        data = bn.get_feature_data(feat)
                        # The data are available, we may abort the search.
                        break
                except (KeyError, OSError, PermissionError):
                    # Basin data not available
                    pass
                except feat_basin.BasinNotAvailableError:
                    # remove the basin from the list
                    # TODO:
                    #  Check whether this has an actual effect. It could be
                    #  that due to some iterative process `self`
                    #  gets re-initialized and we have to go through this
                    #  again.
                    self._basins.remove(bn)
                    warnings.warn(
                        f"Removed unavailable basin {bn} from {self}")
                except BaseException:
                    warnings.warn(f"Could not access {feat} in {self}:\n"
                                  f"{traceback.format_exc()}")
                    pass
        return data

    @staticmethod
    def _apply_scale(a, scale, feat):
        """Helper function for transforming an aray to log-scale

        Parameters
        ----------
        a: np.ndarray
            Input array
        scale: str
            If set to "log", take the logarithm of `a`; if set to
            "linear" return `a` unchanged.
        feat: str
            Feature name (required for debugging)

        Returns
        -------
        b: np.ndarray
            The scaled array

        Notes
        -----
        If the scale is not "linear", then a new array is returned.
        All warnings are suppressed when computing `np.log(a)`, as
        `a` may have negative or nan values.
        """
        if scale == "linear":
            b = a
        elif scale == "log":
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                b = np.log(a)
                if len(w):
                    # Tell the user that the log-transformation issued
                    # a warning.
                    warnings.warn("Invalid values encounterd in np.log "
                                  "while scaling feature '{}'!".format(feat))
        else:
            raise ValueError("`scale` must be either 'linear' or 'log', "
                             + "got '{}'!".format(scale))
        return b

    @staticmethod
    def get_kde_spacing(a, scale="linear", method=kde_methods.bin_width_doane,
                        method_kw=None, feat="undefined", ret_scaled=False):
        """Convenience function for computing the contour spacing

        Parameters
        ----------
        a: ndarray
            feature data
        scale: str
            how the data should be scaled ("log" or "linear")
        method: callable
            KDE method to use (see `kde_methods` submodule)
        method_kw: dict
            keyword arguments to `method`
        feat: str
            feature name for debugging
        ret_scaled: bool
            whether to return the scaled array of `a`
        """
        if method_kw is None:
            method_kw = {}
        # Apply scale (no change for linear scale)
        asc = RTDCBase._apply_scale(a, scale, feat)
        # Apply multiplicator
        acc = method(asc, **method_kw)
        if ret_scaled:
            return acc, asc
        else:
            return acc

    @property
    def _feature_candidates(self):
        """List of feature candidates for this dataset

        Use with caution! Features in this list might not actually
        be available. Always check against `__contains__`.
        """
        feats = list(self._events.keys())
        feats += list(self._usertemp.keys())
        feats += list(AncillaryFeature.feature_names)
        feats += self.features_basin
        feats = sorted(set(feats))
        # exclude non-standard features
        featsv = [ff for ff in feats if dfn.feature_exists(ff)]
        return featsv

    @property
    def _filter(self):
        """return the current filter boolean array"""
        warnings.warn("RTDCBase._filter is deprecated. Please use "
                      + "RTDCBase.filter.all instead.",
                      DeprecationWarning)
        return self.filter.all

    @property
    def _plot_filter(self):
        raise NotImplementedError(
            "RTDCBase._plot_filter has been removed in dclab 0.16.0. "
            + "Please use the output of RTDCBase.downsample_scatter "
            + "with the argument ret_mask instead.")

    @property
    def identifier(self):
        """Unique (unreproducible) identifier"""
        return self._identifier

    @property
    def features(self):
        """All available features"""
        features = []
        for col in self._feature_candidates:
            if col in self:
                features.append(col)
        features.sort()
        return features

    @property
    def features_ancillary(self):
        """All available ancillary features

        This includes all ancillary features, excluding the features
        that are already in `self.features_innate`. This means that
        there may be overlap between `features_ancillary` and e.g.
        `self.features_basin`.

        .. versionadded:: 0.58.0

        """
        features_innate = self.features_innate
        features_ancillary = []
        for feat in AncillaryFeature.feature_names:
            if feat not in features_innate and feat in self:
                features_ancillary.append(feat)
        return sorted(features_ancillary)

    @property
    def features_basin(self):
        """All features accessed via upstream basins from other locations"""
        if self._basins_features is None:
            if self.basins:
                features = []
                for bn in self.basins:
                    if bn.features and set(bn.features) <= set(features):
                        # We already have the features from a different basin.
                        # There might be a basin availability check going on
                        # somewhere, but we are not interested in it.
                        continue
                    if bn.is_available():
                        features += bn.features
                self._basins_features = sorted(set(features))
            else:
                self._basins_features = []
        return self._basins_features

    @property
    def features_innate(self):
        """All features excluding ancillary, basin, or temporary features"""
        innate = [ft for ft in self.features if ft in self._events]
        return innate

    @property
    def features_loaded(self):
        """All features that have been computed

        This includes ancillary features and temporary features.

        Notes
        -----
        Ancillary features that are computationally cheap to compute are
        always included. They are defined in
        :const:`dclab.rtdc_dataset.feat_anc_core.FEATURES_RAPID`.
        """
        features_loaded = self.features_local + self.features_innate
        features_loaded += [f for f in self.features if f in FEATURES_RAPID]
        return sorted(set(features_loaded))

    @property
    def features_local(self):
        """All features that are, with certainty, really fast to access

        Local features is a slimmed down version of `features_loaded`.
        Nothing needs to be computed, not even rapid features
        (:const:`dclab.rtdc_dataset.feat_anc_core.FEATURES_RAPID`).
        And features from remote sources that have not been downloaded
        already are excluded. Ancillary and temporary features that are
        available are included.
        """
        features_local = []
        # Note that the hierarchy format just calls its hparent's
        # `features_local`.
        if hasattr(self._events, "_cached_events"):
            features_local += list(self._events._cached_events.keys())

        if self.format == "hdf5":
            features_local += list(self._events.keys())

        # Get into the basins.
        for bn in self.basins:
            if (bn.basin_format == "hdf5"
                    and bn.basin_type == "file"
                    and bn.is_available()):
                features_local += bn.ds.features_local
            elif bn._ds is not None:
                features_local += bn.ds.features_local

        # If they are here, then we use them:
        features_local += list(self._ancillaries.keys())
        features_local += list(self._usertemp.keys())

        return sorted(set(features_local))

    @property
    def features_scalar(self):
        """All scalar features available"""
        sclr = [ft for ft in self.features if dfn.scalar_feature_exists(ft)]
        return sclr

    @property
    @abc.abstractmethod
    def hash(self):
        """Reproducible dataset hash (defined by derived classes)"""

    def ignore_basins(self, basin_identifiers):
        """Ignore these basin identifiers when looking for features

        This is used to avoid circular basin dependencies.
        """
        self._basins_ignored += basin_identifiers

    def apply_filter(self, force=None):
        """Compute the filters for the dataset"""
        if force is None:
            force = []
        self.filter.update(rtdc_ds=self, force=force)

    def close(self):
        """Close any open files or connections, including basins

        If implemented in a subclass, the subclass must call this
        method via `super`, otherwise basins are not closed. The
        subclass is responsible for closing its specific file handles.
        """
        if self._basins:
            for bn in self._basins:
                bn.close()

    def get_downsampled_scatter(self, xax="area_um", yax="deform",
                                downsample=0, xscale="linear",
                                yscale="linear", remove_invalid=False,
                                ret_mask=False):
        """Downsampling by removing points at dense locations

        Parameters
        ----------
        xax: str
            Identifier for x axis (e.g. "area_um", "aspect", "deform")
        yax: str
            Identifier for y axis
        downsample: int
            Number of points to draw in the down-sampled plot.
            This number is either

            - >=1: exactly downsample to this number by randomly adding
                   or removing points
            - 0  : do not perform downsampling
        xscale: str
            If set to "log", take the logarithm of the x-values before
            performing downsampling. This is useful when data are are
            displayed on a log-scale. Defaults to "linear".
        yscale: str
            See `xscale`.
        remove_invalid: bool
            Remove nan and inf values before downsampling; if set to
            `True`, the actual number of samples returned might be
            smaller than `downsample` due to infinite or nan values
            (e.g. due to logarithmic scaling).
        ret_mask: bool
            If set to `True`, returns a boolean array of length
            `len(self)` where `True` values identify the filtered
            data.

        Returns
        -------
        xnew, xnew: 1d ndarray of lenght `N`
            Filtered data; `N` is either identical to `downsample`
            or smaller (if `remove_invalid==True`)
        mask: 1d boolean array of length `len(RTDCBase)`
            Array for identifying the downsampled data points
        """
        if downsample < 0:
            raise ValueError("`downsample` must be zero or positive!")

        downsample = int(downsample)
        xax = xax.lower()
        yax = yax.lower()

        # Get data
        x = self[xax][self.filter.all]
        y = self[yax][self.filter.all]

        # Apply scale (no change for linear scale)
        xs = RTDCBase._apply_scale(x, xscale, xax)
        ys = RTDCBase._apply_scale(y, yscale, yax)

        _, _, idx = downsampling.downsample_grid(xs, ys,
                                                 samples=downsample,
                                                 remove_invalid=remove_invalid,
                                                 ret_idx=True)

        if ret_mask:
            # Mask is a boolean array of len(self)
            mask = np.zeros(len(self), dtype=bool)
            mids = np.where(self.filter.all)[0]
            mask[mids] = idx
            return x[idx], y[idx], mask
        else:
            return x[idx], y[idx]

    def get_kde_contour(self, xax="area_um", yax="deform", xacc=None,
                        yacc=None, kde_type="histogram", kde_kwargs=None,
                        xscale="linear", yscale="linear"):
        """Evaluate the kernel density estimate for contour plots

        Parameters
        ----------
        xax: str
            Identifier for X axis (e.g. "area_um", "aspect", "deform")
        yax: str
            Identifier for Y axis
        xacc: float
            Contour accuracy in x direction
        yacc: float
            Contour accuracy in y direction
        kde_type: str
            The KDE method to use
        kde_kwargs: dict
            Additional keyword arguments to the KDE method
        xscale: str
            If set to "log", take the logarithm of the x-values before
            computing the KDE. This is useful when data are
            displayed on a log-scale. Defaults to "linear".
        yscale: str
            See `xscale`.

        Returns
        -------
        X, Y, Z : coordinates
            The kernel density Z evaluated on a rectangular grid (X,Y).
        """
        if kde_kwargs is None:
            kde_kwargs = {}
        xax = xax.lower()
        yax = yax.lower()
        kde_type = kde_type.lower()
        if kde_type not in kde_methods.methods:
            raise ValueError("Not a valid kde type: {}!".format(kde_type))

        # Get data
        x = self[xax][self.filter.all]
        y = self[yax][self.filter.all]

        xacc_sc, xs = RTDCBase.get_kde_spacing(
            a=x,
            feat=xax,
            scale=xscale,
            method=kde_methods.bin_width_doane,
            ret_scaled=True)

        yacc_sc, ys = RTDCBase.get_kde_spacing(
            a=y,
            feat=yax,
            scale=yscale,
            method=kde_methods.bin_width_doane,
            ret_scaled=True)

        if xacc is None or xacc == 0:
            xacc = xacc_sc / 5

        if yacc is None or yacc == 0:
            yacc = yacc_sc / 5

        # Ignore infs and nans
        bad = kde_methods.get_bad_vals(xs, ys)
        xc = xs[~bad]
        yc = ys[~bad]

        xnum = int(np.ceil((xc.max() - xc.min()) / xacc))
        ynum = int(np.ceil((yc.max() - yc.min()) / yacc))

        xlin = np.linspace(xc.min(), xc.max(), xnum, endpoint=True)
        ylin = np.linspace(yc.min(), yc.max(), ynum, endpoint=True)

        xmesh, ymesh = np.meshgrid(xlin, ylin, indexing="ij")

        kde_fct = kde_methods.methods[kde_type]
        if len(x):
            density = kde_fct(events_x=xs, events_y=ys,
                              xout=xmesh, yout=ymesh,
                              **kde_kwargs)
        else:
            density = np.array([])

        # Convert mesh back to linear scale if applicable
        if xscale == "log":
            xmesh = np.exp(xmesh)
        if yscale == "log":
            ymesh = np.exp(ymesh)

        return xmesh, ymesh, density

    def get_kde_scatter(self, xax="area_um", yax="deform", positions=None,
                        kde_type="histogram", kde_kwargs=None, xscale="linear",
                        yscale="linear"):
        """Evaluate the kernel density estimate for scatter plots

        Parameters
        ----------
        xax: str
            Identifier for X axis (e.g. "area_um", "aspect", "deform")
        yax: str
            Identifier for Y axis
        positions: list of two 1d ndarrays or ndarray of shape (2, N)
            The positions where the KDE will be computed. Note that
            the KDE estimate is computed from the points that
            are set in `self.filter.all`.
        kde_type: str
            The KDE method to use, see :const:`.kde_methods.methods`
        kde_kwargs: dict
            Additional keyword arguments to the KDE method
        xscale: str
            If set to "log", take the logarithm of the x-values before
            computing the KDE. This is useful when data are are
            displayed on a log-scale. Defaults to "linear".
        yscale: str
            See `xscale`.

        Returns
        -------
        density : 1d ndarray
            The kernel density evaluated for the filtered data points.
        """
        if kde_kwargs is None:
            kde_kwargs = {}
        xax = xax.lower()
        yax = yax.lower()
        kde_type = kde_type.lower()
        if kde_type not in kde_methods.methods:
            raise ValueError("Not a valid kde type: {}!".format(kde_type))

        # Get data
        x = self[xax][self.filter.all]
        y = self[yax][self.filter.all]

        # Apply scale (no change for linear scale)
        xs = RTDCBase._apply_scale(x, xscale, xax)
        ys = RTDCBase._apply_scale(y, yscale, yax)

        if positions is None:
            posx = None
            posy = None
        else:
            posx = RTDCBase._apply_scale(positions[0], xscale, xax)
            posy = RTDCBase._apply_scale(positions[1], yscale, yax)

        kde_fct = kde_methods.methods[kde_type]
        if len(x):
            density = kde_fct(events_x=xs, events_y=ys,
                              xout=posx, yout=posy,
                              **kde_kwargs)
        else:
            density = np.array([])

        return density

    def basins_get_dicts(self):
        """Return the list of dictionaries describing the dataset's basins"""
        # Only implement this for classes that support this
        return []

    def basins_retrieve(self):
        """Load all basins available

        .. versionadded:: 0.54.0

        In dclab 0.51.0, we introduced basins, a simple way of combining
        HDF5-based datasets (including the :class:`.HDF5_S3` format).
        The idea is to be able to store parts of the dataset
        (e.g. images) in a separate file that could then be located
        someplace else (e.g. an S3 object store).

        If an RT-DC file has "basins" defined, then these are sought out and
        made available via the `features_basin` property.

        .. versionchanged:: 0.57.5

            "file"-type basins are only available for subclasses that
            set the `_local_basins_allowed` attribute to True.
        """
        basins = []
        bc = feat_basin.get_basin_classes()
        # Sort basins according to priority
        bdicts_srt = sorted(self.basins_get_dicts(),
                            key=feat_basin.basin_priority_sorted_key)
        # complement basin "key"s (we do the same in writer)
        for bdict in bdicts_srt:
            if "key" not in bdict:
                b_dat = json.dumps(bdict, indent=2, sort_keys=True).split("\n")
                bdict["key"] = hashobj(b_dat)

        bd_keys = [bd["key"] for bd in bdicts_srt]
        bd_keys += self._basins_ignored
        for bdict in bdicts_srt:
            if bdict["format"] not in bc:
                warnings.warn(f"Encountered unsupported basin "
                              f"format '{bdict['format']}'!")
                continue
            if bdict["key"] in self._basins_ignored:
                warnings.warn(
                    f"Encountered cyclic basin dependency '{bdict['key']}'",
                    feat_basin.CyclicBasinDependencyFoundWarning)
                continue

            # Basin initialization keyword arguments
            kwargs = {
                "name": bdict.get("name"),
                "description": bdict.get("description"),
                # Honor features intended by basin creator.
                "features": bdict.get("features"),
                # Which mapping we are using ("same", "basinmap1", ...)
                "mapping": bdict.get("mapping", "same"),
                # For non-identical mapping ("basinmap1", etc.), we
                # need the referring dataset.
                "mapping_referrer": self,
                # Make sure the measurement identifier is checked.
                "measurement_identifier": self.get_measurement_identifier(),
                # allow to ignore basins
                "ignored_basins": bd_keys,
                # basin key
                "key": bdict["key"],
            }

            # Check whether this basin is supported and exists
            if bdict["type"] == "internal":
                b_cls = bc[bdict["format"]]
                bna = b_cls(bdict["paths"][0], **kwargs)
                # In contrast to file-type basins, we just add all remote
                # basins without checking first. We do not check for
                # the availability of remote basins, because they could
                # be temporarily inaccessible (unstable network connection)
                # and because checking the availability of remote basins
                # normally takes a lot of time.
                basins.append(bna)
            elif bdict["type"] == "file":
                if not self._local_basins_allowed:
                    warnings.warn(f"Basin type 'file' not allowed for format "
                                  f"'{self.format}'")
                    # stop processing this basin
                    continue
                p_paths = list(bdict["paths"])
                # translate Windows and Unix relative paths
                for pi in list(p_paths):  # [sic] create a copy of the list
                    if pi.count(".."):
                        if pi[2:].count("/") and os.path.sep == r"\\":
                            # Windows
                            p_paths.append(pi.replace("/", r"\\"))
                        elif pi[2:].count(r"\\") and os.path.sep == "/":
                            # Unix
                            p_paths.append(pi.replace(r"\\", "/"))
                # perform the actual check
                for pp in p_paths:
                    pp = pathlib.Path(pp)
                    # Instantiate the proper basin class
                    b_cls = bc[bdict["format"]]
                    # Try absolute path
                    bna = b_cls(pp, **kwargs)
                    if bna.verify_basin():
                        basins.append(bna)
                        break
                    # Try relative path
                    this_path = pathlib.Path(self.path)
                    if this_path.exists():
                        # Insert relative path
                        bnr = b_cls(this_path.parent / pp, **kwargs)
                        if bnr.verify_basin():
                            basins.append(bnr)
                            break
            elif bdict["type"] == "remote":
                for url in bdict["urls"]:
                    # Instantiate the proper basin class
                    b_cls = bc[bdict["format"]]
                    bna = b_cls(url, **kwargs)
                    # In contrast to file-type basins, we just add all remote
                    # basins without checking first. We do not check for
                    # the availability of remote basins, because they could
                    # be temporarily inaccessible (unstable network connection)
                    # and because checking the availability of remote basins
                    # normally takes a lot of time.
                    basins.append(bna)
            else:
                warnings.warn(
                    f"Encountered unsupported basin type '{bdict['type']}'!")
        return basins

    def get_measurement_identifier(self):
        """Return a unique measurement identifier

        Return the [experiment]:"run identifier" configuration feat, if it
        exists. Otherwise, return the MD5 sum computed from the measurement
        time, date, and setup identifier.

        Returns `None` if no identifier could be found or computed.

        .. versionadded:: 0.51.0

        """
        identifier = self.config.get("experiment", {}).get("run identifier",
                                                           None)
        if identifier is None:
            time = self.config.get("experiment", {}).get("time", None)
            date = self.config.get("experiment", {}).get("date", None)
            sid = self.config.get("setup", {}).get("identifier", None)
            if None not in [time, date, sid]:
                # only compute an identifier if all of the above are defined.
                hasher = hashlib.md5(f"{time}_{date}_{sid}".encode("utf-8"))
                identifier = str(uuid.UUID(hex=hasher.hexdigest()))
        return identifier

    def polygon_filter_add(self, filt):
        """Associate a Polygon Filter with this instance

        Parameters
        ----------
        filt: int or instance of `PolygonFilter`
            The polygon filter to add
        """
        self._assert_filter()  # [sic] initialize the filter if not done yet
        if not isinstance(filt, (PolygonFilter, int, float)):
            msg = "`filt` must be a number or instance of PolygonFilter!"
            raise ValueError(msg)

        if isinstance(filt, PolygonFilter):
            uid = filt.unique_id
        else:
            uid = int(filt)
        # append item
        self.config["filtering"]["polygon filters"].append(uid)

    def polygon_filter_rm(self, filt):
        """Remove a polygon filter from this instance

        Parameters
        ----------
        filt: int or instance of `PolygonFilter`
            The polygon filter to remove
        """
        if not isinstance(filt, (PolygonFilter, int, float)):
            msg = "`filt` must be a number or instance of PolygonFilter!"
            raise ValueError(msg)

        if isinstance(filt, PolygonFilter):
            uid = filt.unique_id
        else:
            uid = int(filt)
        # remove item
        self.config["filtering"]["polygon filters"].remove(uid)

    def reset_filter(self):
        """Reset the current filter"""
        # reset filter instance
        self.filter.reset()
        # reset configuration
        # remember hierarchy parent
        hp = self.config["filtering"]["hierarchy parent"]
        self.config._init_default_filter_values()
        self.config["filtering"]["hierarchy parent"] = hp



# flake8: noqa: F401
from .base import RTDC_Hierarchy
from .events import (
    ChildTrace, ChildTraceItem, ChildScalar, ChildContour, ChildNDArray,
    ChildBase
)
from .hfilter import HierarchyFilter, HierarchyFilterError
from .mapper import (
    map_indices_child2parent, map_indices_child2root,
    map_indices_root2child, map_indices_parent2child
)



"""Verify dataset integrity"""
import argparse
import pathlib
import sys

from ..rtdc_dataset import check_dataset, fmt_tdms
from .._version import version

from . import common


def verify_dataset(path_in=None):
    """Perform checks on experimental datasets"""
    if path_in is None:
        parser = verify_dataset_parser()
        args = parser.parse_args()
        path_in = pathlib.Path(args.path).resolve()
    if not path_in.exists():
        common.print_violation(f"File not found: {path_in}")
        return sys.exit(4)
    else:
        common.print_info(f"Checking {path_in}")
    # The exit status of this script. Non-zero exit status means:
    # 1: alerts
    # 2: violations
    # 3: alerts and violations
    # 4: other error
    exit_status = 4
    try:
        viol, aler, info = check_dataset(path_in)
    except fmt_tdms.InvalidTDMSFileFormatError:
        common.print_violation("Invalid tdms file format!")
    except fmt_tdms.IncompleteTDMSFileFormatError:
        common.print_violation("Incomplete dataset!")
    except fmt_tdms.ContourIndexingError:
        common.print_violation("Invalid contour data!")
    except fmt_tdms.InvalidVideoFileError:
        common.print_violation("Invalid image data!")
    except BaseException as e:
        common.print_violation(f"{e.__class__.__name__}: {', '.join(e.args)}")
    else:
        for inf in info:
            common.print_info(inf)
        for ale in aler:
            common.print_alert(ale)
        for vio in viol:
            common.print_violation(vio)
        common.print_info(f"Check Complete: {len(viol)} violations and "
                          + f"{len(aler)} alerts")
        if aler and viol:
            exit_status = 3
        elif aler:
            exit_status = 1
        elif viol:
            exit_status = 2
        else:
            # everything is ok
            exit_status = 0
    finally:
        # return sys.exit for testing (monkeypatched)
        return sys.exit(exit_status)


def verify_dataset_parser():
    descr = "Check experimental datasets for completeness. This command " \
            + "is used e.g. to enforce data integrity with Shape-In. The " \
            + "following exit codes are defined: ``0: valid dataset``, " \
            + "``1: alerts encountered``, ``2: violations encountered``, " \
            + "``3: alerts and violations``, ``4: other error``."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('path', metavar='PATH', type=str,
                        help='Path to experimental dataset')
    parser.add_argument('--version', action='version',
                        version=f'dclab-verify-dataset {version}')
    return parser



import pathlib

import pytest

from dclab.rtdc_dataset import fmt_dcor
from dclab.rtdc_dataset.fmt_dcor import access_token

datapath = pathlib.Path(__file__).parent / "data"


def test_get_api_key():
    api_key = access_token.get_api_key(
        datapath / "example_access_token.dcor-access", "42")
    assert api_key == "7c0c7203-4e25-4b14-a118-553c496a7a52"


def test_get_certificate():
    cert = access_token.get_certificate(
        datapath / "example_access_token.dcor-access", "42").decode()
    assert "yTCCBLGgAwIBAgIUSrQD5LuXBSUtn41PeGDqP9XPbVIwDQYJKoZIhvcNA" in cert


def test_get_hostname():
    hostname = access_token.get_hostname(
        datapath / "example_access_token.dcor-access", "42")
    assert hostname == "dcor.example.com"


# feeling very confident about https://github.com/pytest-dev/pytest/pull/9871
@pytest.mark.xfail(pytest.version_tuple < (7, 2, 0),
                   reason="Requires pytest PR #9871 when run with coverage")
def test_store_and_get_certificate(tmp_path):
    cert = access_token.get_certificate(
        datapath / "example_access_token.dcor-access", "42")
    expect_path = tmp_path / "dcor.example.com.cert"
    expect_path.write_bytes(cert)
    fmt_dcor.DCOR_CERTS_SEARCH_PATHS.append(tmp_path)
    try:
        cert_path = fmt_dcor.get_server_cert_path("dcor.example.com")
    except BaseException:
        raise
    else:
        assert str(cert_path) == str(expect_path)
    finally:
        fmt_dcor.DCOR_CERTS_SEARCH_PATHS.clear()



"""Concatenate .rtdc files"""
from __future__ import annotations

import argparse
import pathlib
import time
from typing import Dict, List
import warnings

import hdf5plugin
import numpy as np

from ..rtdc_dataset import new_dataset, RTDCWriter
from .. import definitions as dfn
from .._version import version

from . import common


class FeatureSetNotIdenticalJoinWarning(UserWarning):
    pass


def join(
        paths_in: List[str | pathlib.Path] = None,
        path_out: str | pathlib.Path = None,
        metadata: Dict = None,
        ret_path: bool = False,
        ):
    """Join multiple RT-DC measurements into a single .rtdc file

    Parameters
    ----------
    paths_in: list of paths
        input paths to join
    path_out: str or pathlib.Path
        output path
    metadata: dict
        optional metadata dictionary (configuration dict) to store
        in the output file
    ret_path: bool
        whether to return the output path

    Returns
    -------
    path_out: pathlib.Path (optional)
        output path (with corrected path suffix if applicable)

    Notes
    -----
    The first input file defines the metadata written to the output
    file. Only features that are present in all input files are written
    to the output file.
    """
    cmp_kw = hdf5plugin.Zstd(clevel=5)
    if metadata is None:
        metadata = {"experiment": {"run index": 1}}
    if path_out is None or paths_in is None:
        parser = join_parser()
        args = parser.parse_args()
        paths_in = args.input
        path_out = args.output

    if len(paths_in) < 2:
        raise ValueError("At least two input files must be specified!")

    paths_in, path_out, path_temp = common.setup_task_paths(
        paths_in, path_out, allowed_input_suffixes=[".rtdc", ".tdms"])

    # Order input files by date
    key_paths = []
    for pp in paths_in:
        with new_dataset(pp) as dsa:
            # sorting key
            key = "_".join([dsa.config["experiment"]["date"],
                            dsa.config["experiment"]["time"],
                            str(dsa.config["experiment"]["run index"])
                            ])
            key_paths.append((key, pp))
    sorted_paths = [p[1] for p in sorted(key_paths, key=lambda x: x[0])]

    logs = {"dclab-join": common.get_command_log(paths=sorted_paths)}

    # Determine temporal offsets
    t_offsets = np.zeros(len(sorted_paths), dtype=np.float64)
    for ii, pp in enumerate(sorted_paths):
        with new_dataset(pp) as dsb:
            etime = dsb.config["experiment"]["time"]
            st = time.strptime(dsb.config["experiment"]["date"]
                               + etime[:8],
                               "%Y-%m-%d%H:%M:%S")
            t_offsets[ii] = time.mktime(st)
            if len(etime) > 8:
                # floating point time stored as well (HH:MM:SS.SS)
                t_offsets[ii] += float(etime[8:])
    t_offsets -= t_offsets[0]

    # Determine features to export (based on first file)
    with warnings.catch_warnings(record=True) as w:
        # Catch all FeatureSetNotIdenticalJoinWarnings
        warnings.simplefilter("ignore")
        warnings.simplefilter("always",
                              category=FeatureSetNotIdenticalJoinWarning)
        features = None
        for pp in sorted_paths:
            with new_dataset(pp) as dsc:
                # features present
                if features is None:
                    # The initial features are the innate features of the
                    # first file (sorted by time). If we didn't use the innate
                    # features, then the resulting file might become large
                    # (e.g. if we included ancillary features).
                    features = sorted(dsc.features_innate)
                else:
                    # Remove features from the feature list, if it is not in
                    # this dataset, or cannot be computed on-the-fly.
                    for feat in features:
                        if feat not in dsc.features:
                            features.remove(feat)
                            warnings.warn(
                                f"Excluding feature '{feat}', because "
                                + f"it is not present in '{pp}'!",
                                FeatureSetNotIdenticalJoinWarning)
                    # Warn the user if this dataset has an innate feature that
                    # is being ignored, because it is not an innate feature of
                    # the first dataset.
                    for feat in dsc.features_innate:
                        if feat not in features:
                            warnings.warn(
                                f"Ignoring feature '{feat}' in '{pp}', "
                                + "because it is not present in the "
                                + "other files being joined!",
                                FeatureSetNotIdenticalJoinWarning)
        if w:
            logs["dclab-join-feature-warnings"] = common.assemble_warnings(w)

    # Create initial output file
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        with new_dataset(sorted_paths[0]) as ds0:
            ds0.export.hdf5(path=path_temp,
                            features=features,
                            filtered=False,
                            override=True,
                            logs=True,
                            tables=True,
                            basins=False,
                            meta_prefix="src-#1_",
                            compression_kwargs=cmp_kw)
            # store configuration
            cfg0 = ds0.config.tostring(
                sections=dfn.CFG_METADATA).split("\n")
        if w:
            logs["dclab-join-warnings-#1"] = common.assemble_warnings(w)

    with RTDCWriter(path_temp, compression_kwargs=cmp_kw) as hw:
        # store configuration of first dataset
        hw.store_log(name="src-#1_cfg", lines=cfg0)
        ii = 1
        # Append data from other files
        for pi, ti in zip(sorted_paths[1:], t_offsets[1:]):
            ii += 1  # we start with the second dataset
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                meta_key = f"src-#{ii}"
                meta_prefix = meta_key + "_"
                with new_dataset(pi) as dsi:
                    for feat in features:
                        if feat == "time":
                            # handle time offset
                            fdata = dsi["time"] + ti
                        elif feat == "frame":
                            # handle frame offset
                            fr = dsi.config["imaging"]["frame rate"]
                            fdata = (np.array(dsi["frame"], dtype=np.uint64)
                                     + np.uint64(round(ti * fr)))
                        elif feat == "index_online":
                            if "events/index_online" in hw.h5file:
                                # index_online is usually larger than index
                                ido0 = hw.h5file["events/index_online"][-1] + 1
                            else:
                                ido0 = 0
                            fdata = dsi["index_online"] + ido0
                        else:
                            fdata = dsi[feat]
                        hw.store_feature(feat=feat, data=fdata)
                    # store logs
                    for log in dsi.logs:
                        hw.store_log(name=meta_prefix + log,
                                     lines=dsi.logs[log])
                    # store tables
                    for tab in dsi.tables:
                        hw.store_table(name=meta_prefix + tab,
                                       cmp_array=dsi.tables[tab])
                    # store configuration
                    cfg = dsi.config.tostring(
                        sections=dfn.CFG_METADATA).split("\n")
                    hw.store_log(name=f"{meta_key}_cfg",
                                 lines=cfg)
                if w:
                    hw.store_log(name=f"dclab-join-warnings-#{ii}",
                                 lines=common.assemble_warnings(w))

        # Write logs and missing meta data
        for name in logs:
            hw.store_log(name, logs[name])
        hw.store_metadata(metadata)

    # Finally, rename temp to out
    path_temp.rename(path_out)
    if ret_path:
        return path_out


def join_parser():
    descr = "Join two or more RT-DC measurements. This will produce " \
            + "one larger .rtdc file. The meta data of the dataset " \
            + "that was recorded earliest will be used in the output " \
            + "file. Please only join datasets that were recorded " \
            + "in the same measurement run."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('input', metavar="INPUT", nargs="*", type=str,
                        help='Input paths (.tdms or .rtdc files)')
    required_named = parser.add_argument_group('required named arguments')
    required_named.add_argument('-o', '--output', metavar="OUTPUT", type=str,
                                help='Output path (.rtdc file)', required=True)
    parser.add_argument('--version', action='version',
                        version=f'dclab-join {version}')
    return parser



import h5py
import numpy as np
import pytest

import dclab

from helper_methods import retrieve_data


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_brightness_bc():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    ds = dclab.new_dataset(path)
    # sanity checks
    assert "bright_bc_avg" not in ds.features_innate
    assert "bright_bc_sd" not in ds.features_innate
    # ignore first event (no image data)
    for ii in range(1, len(ds)):
        image_corr = np.array(ds["image"][ii], dtype=int) - ds["image_bg"][ii]
        mask = ds["mask"][ii]
        assert np.max(np.abs(image_corr)) < 50, "sanity check"
        assert ds["bright_bc_avg"][ii] == np.mean(image_corr[mask])
        assert ds["bright_bc_sd"][ii] == np.std(image_corr[mask])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_brightness_bc_with_offset():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    with h5py.File(path, "a") as h5:
        bg_off = np.linspace(0, 1, len(h5["events/deform"]))
        h5["events/bg_off"] = bg_off

    ds = dclab.new_dataset(path)
    # sanity checks
    assert "bright_bc_avg" not in ds.features_innate
    assert "bright_bc_sd" not in ds.features_innate
    # ignore first event (no image data)
    for ii in range(1, len(ds)):
        image_bg = ds["image_bg"][ii] + bg_off[ii]
        image_corr = np.array(ds["image"][ii], dtype=int) - image_bg
        mask = ds["mask"][ii]
        assert np.max(np.abs(image_corr)) < 50, "sanity check"
        assert ds["bright_bc_avg"][ii] == np.mean(image_corr[mask])
        assert ds["bright_bc_sd"][ii] == np.std(image_corr[mask])



"""Overview plots for the Young's modulus look-up table

The Young's modulus is corrected for pixelation (pixel size 0.34 µm).
The temperature is set to 23°C.
The 20µm plot is used in the dclab documentation. The others
are used in the Shape-Out documentation.
"""
from dclab.features import emodulus
import numpy as np
import seaborn as sns
import matplotlib.pylab as plt
from matplotlib import cm, colors
from matplotlib.ticker import MultipleLocator

sns.set_style("whitegrid")

media = ["CellCarrier", "CellCarrier B", "water"]
matrix = {  # channel width: flow rates
    15: [0.016, 0.032, 0.048],
    20: [0.040, 0.080, 0.120],
    30: [0.160, 0.240, 0.320],
    40: [0.320, 0.400, 0.600],
}
ylim = (0, 0.2)
xlim = {  # channel width: xmax
    15: (10, 200),
    20: (20, 350),
    30: (50, 800),
    40: (120, 1400),
}

deform_max = 0.17
grid_size = 1000
emin = 4e-2
emax = 2e1
norm = colors.LogNorm(vmin=emin, vmax=emax)
cmap = "jet"


for lut_name in [
    "LE-2D-FEM-19",
    "HE-2D-FEM-22",
    "HE-3D-FEM-22"
]:

    for channel_width in matrix:
        flow_rates = matrix[channel_width]
        # setup figure
        fig, axes = plt.subplots(nrows=3, ncols=3, sharex=True, sharey=True,
                                 gridspec_kw={"left": 0.12,
                                              "right": .85,
                                              "wspace": .1,
                                              "hspace": .15,
                                              },
                                 figsize=(7.5, 6.3))
        fig.suptitle("{} µm channel".format(channel_width))
        axes[2][1].set_xlabel("Area [µm²]")
        axes[1][0].set_ylabel("Deformation")
        # column label (flow rate)
        for ii in range(3):
            axes[0][ii].set_title("{} µL/s".format(flow_rates[ii]),
                                  fontsize=10, fontweight="bold")
        # row label (medium)
        for jj in range(3):
            pos = axes[jj][0].get_position()
            x = pos.x0 - .11
            y = pos.y0 + pos.height/2
            fig.text(x, y, media[jj], ha='left', va='center', rotation=90,
                     fontsize=10, fontweight="bold")

        # plot the LUTs channel-wise
        lut_base, lut_meta = emodulus.load_lut(lut_name)
        deform = np.linspace(lut_base[:, 1].min(), lut_base[:, 1].max(),
                             grid_size, endpoint=True)
        area_um = np.linspace(lut_base[:, 0].min(), lut_base[:, 0].max(),
                              grid_size, endpoint=True)
        for jj, medium in enumerate(media):  # rows
            for ii, flow_rate in enumerate(matrix[channel_width]):  # columns
                area_um_scaled = emodulus.scale_linear.scale_area_um(
                    area_um=area_um,
                    channel_width_in=lut_meta["channel_width"],
                    channel_width_out=channel_width)
                aa, dd = np.meshgrid(area_um_scaled, deform)
                emod = emodulus.get_emodulus(area_um=aa,
                                             deform=dd,
                                             medium=medium,
                                             channel_width=channel_width,
                                             flow_rate=flow_rate,
                                             px_um=.34,
                                             temperature=23.0,
                                             visc_model="buyukurganci-2022",
                                             lut_data=lut_name)
                ax = axes[jj][ii]
                ax.set_aspect("auto")
                pc = ax.pcolormesh(aa, dd, emod, shading="nearest", cmap=cmap,
                                   norm=norm, zorder=11)
                ax.grid(zorder=10)
                # ticks every .5
                ax.yaxis.set_major_locator(MultipleLocator(0.05))

        axes[2][2].set_ylim(*ylim)
        axes[2][2].set_xlim(*xlim[channel_width])
        axes[2][2].grid(True, which="minor", zorder=10)

        # add colorbar
        cax = plt.axes([0.87, 0.1, 0.02, 0.8])

        mappable = cm.ScalarMappable(norm=norm, cmap=cmap)
        plt.colorbar(pc, orientation="vertical", cax=cax, extend="both",
                     label=f"apparent Young's modulus ({lut_name}) [kPa]")

        plt.savefig("_emodulus_{}um_{}.png".format(channel_width, lut_name),
                    dpi=150)
        plt.close()



import logging
import os
import pathlib
import shutil
import subprocess as sp

logger = logging.getLogger(__name__)

_has_lme4 = None
_has_r = None


class CommandFailedError(BaseException):
    """Used when `run_command` encounters an error"""
    pass


class RNotFoundError(BaseException):
    pass


def get_r_path():
    """Return the path of the R executable"""
    # Maybe the user set the executable already?
    r_exec = os.environ.get("R_EXEC")
    if r_exec is not None:
        r_exec = pathlib.Path(r_exec)
        if r_exec.is_file():
            return r_exec

    # Try to get the executable using which
    r_exec = shutil.which("R")
    if r_exec is not None:
        r_exec = pathlib.Path(r_exec)
        return r_exec

    # Try to determine the path to the executable from R_HOME
    r_home = os.environ.get("R_HOME")
    if r_home and not pathlib.Path(r_home).is_dir():
        logger.warning(f"R_HOME Directory does not exist: {r_home}")
        r_home = None

    if r_home is None:
        raise RNotFoundError(
            "Cannot find R, please set the `R_HOME` environment variable "
            "or use `set_r_path`.")

    r_home = pathlib.Path(r_home)

    # search for the R executable
    for rr in [
        r_home / "bin" / "R",
        r_home / "bin" / "x64" / "R",
    ]:
        if rr.is_file():
            return rr
        rr_win = rr.with_name("R.exe")
        if rr_win.is_file():
            return rr_win
    else:
        raise RNotFoundError(
            f"Could not find R binary in '{r_home}'")


def get_r_script_path():
    """Return the path to the Rscript executable"""
    return get_r_path().with_name("Rscript")


def get_r_version():
    """Return the full R version string"""
    require_r()
    cmd = (str(get_r_path()), "--version")
    logger.debug(f"Looking for R version with: {' '.join(cmd)}")
    r_version = run_command(
        cmd,
        env={"R_LIBS_USER": os.environ.get("R_LIBS_USER", "")},
    )
    r_version = r_version.split(os.linesep)
    if r_version[0].startswith("WARNING"):
        r_version = r_version[1]
    else:
        r_version = r_version[0]
    logger.info(f"R version found: {r_version}")
    # get the actual version string
    if r_version.startswith("R version "):
        r_version = r_version.split(" ", 2)[2]
    return r_version.strip()


def has_lme4():
    """Return True if the lme4 package is installed"""
    global _has_lme4
    if _has_lme4:
        return True
    require_r()
    for pkg in ["lme4", "statmod", "nloptr"]:
        res = run_command(
            (str(get_r_path()), "-q", "-e", f"system.file(package='{pkg}')"),
            env={"R_LIBS_USER": os.environ.get("R_LIBS_USER", "")},
            )
        if not res.split("[1]")[1].count(pkg):
            avail = False
            break
    else:
        avail = _has_lme4 = True
    return avail


def has_r():
    """Return True if R is available"""
    global _has_r
    if _has_r:
        return True
    try:
        hasr = get_r_path().is_file()
    except RNotFoundError:
        hasr = False
    if hasr:
        _has_r = True
    return hasr


def require_lme4():
    """Install the lme4 package (if not already installed)

    Besides ``lme4``, this also installs ``nloptr`` and ``statmod``.
    The packages are installed to the user data directory
    given in :const:`lib_path` from the http://cran.rstudio.org mirror.
    """
    install_command = ("install.packages("
                       "c('statmod','nloptr','lme4'),"
                       "repos='http://cran.rstudio.org'"
                       ")"
                       )
    require_r()
    if not has_lme4():
        run_command(cmd=(get_r_path(), "-e", install_command),
                    env={"R_LIBS_USER": os.environ.get("R_LIBS_USER", "")},
                    )


def require_r():
    """Make sure R is installed an R HOME is set"""
    if not has_r():
        raise RNotFoundError("Cannot find R, please set its path with the "
                             "`set_r_path` function or set the `RHOME` "
                             "environment variable.")


def run_command(cmd, **kwargs):
    """Run a command via subprocess"""
    if hasattr(sp, "STARTUPINFO"):
        # On Windows, subprocess calls will pop up a command window by
        # default when run from Pyinstaller with the ``--noconsole``
        # option. Avoid this distraction.
        si = sp.STARTUPINFO()
        si.dwFlags |= sp.STARTF_USESHOWWINDOW
        # Windows doesn't search the path by default. Pass it an
        # environment so it will.
        env = os.environ
    else:
        si = None
        env = None

    kwargs.setdefault("text", True)
    kwargs.setdefault("stderr", sp.STDOUT)
    if env is not None:
        if "env" in kwargs:
            env.update(kwargs.pop("env"))
        kwargs["env"] = env
    kwargs["startupinfo"] = si

    # Convert paths to strings
    cmd = [str(cc) for cc in cmd]

    try:
        tmp = sp.check_output(cmd, **kwargs)
    except sp.CalledProcessError as e:
        raise CommandFailedError(f"The command '{' '.join(cmd)}' failed with "
                                 f"exit code {e.returncode}: {e.output}")

    return tmp.strip()


def set_r_lib_path(r_lib_path):
    """Add given directory to the R_LIBS_USER environment variable"""
    paths = os.environ.get("R_LIBS_USER", "").split(os.pathsep)
    paths = [p for p in paths if p]
    paths.append(str(r_lib_path).strip())
    os.environ["R_LIBS_USER"] = os.pathsep.join(list(set(paths)))


def set_r_path(r_path):
    """Set the path of the R executable/binary"""
    tmp = run_command((str(r_path), "RHOME"))

    r_home = tmp.split(os.linesep)
    if r_home[0].startswith("WARNING"):
        res = r_home[1]
    else:
        res = r_home[0].strip()
    os.environ["R_HOME"] = res
    os.environ["R_EXEC"] = str(pathlib.Path(r_path).resolve())



class ContourIndexingError(BaseException):
    """Use when contour indices are confused"""
    pass


class IncompleteTDMSFileFormatError(BaseException):
    """Use for incomplete dataset (e.g. missing para.ini)"""


class InvalidTDMSFileFormatError(BaseException):
    """Use for invalid tdms files (e.g. unknown columns)"""
    pass


class InvalidVideoFileError(BaseException):
    """Used for bad video files"""
    pass


class CorruptFrameWarning(UserWarning):
    """Video frame corrupt or missing"""
    pass


class InitialFrameMissingWarning(CorruptFrameWarning):
    """Initial frame of video is missing"""
    pass


class SlowVideoWarning(UserWarning):
    """Getting video data will be slow"""
    pass


class MultipleSamplesPerEventFound(UserWarning):
    """Ambiguities in trace data"""
    pass



"""LUT hooks

This submodule contains manually-designed hooks for the different
LUTs used in dclab. Interpolating data on an irregular grid has its
pitfalls. Extracting isoelastics is tricky at the boundaries of the
simulation support. The main idea that is followed here is to
extrapolated isoelastics beyond the convec hull of the original
LUT to then get cleaner isoelastics up until the convex hull.
"""

from . import le_2d_fem_19
from . import he_2d_fem_22
from . import he_3d_fem_22


lut_hooks = {
    "HE-2D-FEM-22": he_2d_fem_22,
    "HE-3D-FEM-22": he_3d_fem_22,
    "LE-2D-FEM-19": le_2d_fem_19,
}



"""Content-based downsampling of ndarrays"""
import numpy as np
cimport numpy as cnp

from .cached import Cache

# We are using the numpy array API
cnp.import_array()
ctypedef cnp.uint8_t uint8
ctypedef cnp.uint32_t uint32
ctypedef cnp.int64_t int64

def downsample_rand(a, samples, remove_invalid=False, ret_idx=False):
    """Downsampling by randomly removing points

    Parameters
    ----------
    a: 1d ndarray
        The input array to downsample
    samples: int
        The desired number of samples
    remove_invalid: bool
        Remove nan and inf values before downsampling
    ret_idx: bool
        Also return a boolean array that corresponds to the
        downsampled indices in `a`.

    Returns
    -------
    dsa: 1d ndarray of size `samples`
        The pseudo-randomly downsampled array `a`
    idx: 1d boolean array with same shape as `a`
        Only returned if `ret_idx` is True.
        A boolean array such that `a[idx] == dsa`
    """
    # fixed random state for this method
    rs = np.random.RandomState(seed=47).get_state()
    np.random.set_state(rs)

    cdef uint32 samples_int = np.uint32(samples)

    if remove_invalid:
        # slice out nans and infs
        bad = np.isnan(a) | np.isinf(a)
        pool = a[~bad]
    else:
        pool = a

    if samples_int and (samples_int < pool.shape[0]):
        keep = np.zeros_like(pool, dtype=bool)
        keep_ids = np.random.choice(np.arange(pool.size),
                                    size=samples_int,
                                    replace=False)
        keep[keep_ids] = True
        dsa = pool[keep]
    else:
        keep = np.ones_like(pool, dtype=bool)
        dsa = pool

    if remove_invalid:
        # translate the kept values back to the original array
        idx = np.zeros(a.size, dtype=bool)
        idx[~bad] = keep
    else:
        idx = keep

    if ret_idx:
        return dsa, idx
    else:
        return dsa


@Cache
def downsample_grid(a, b, samples, remove_invalid=False, ret_idx=False):
    """Content-based downsampling for faster visualization

    The arrays `a` and `b` make up a 2D scatter plot with high
    and low density values. This method takes out points at
    indices with high density.

    Parameters
    ----------
    a, b: 1d ndarrays
        The input arrays to downsample
    samples: int
        The desired number of samples
    remove_invalid: bool
        Remove nan and inf values before downsampling; if set to
        `True`, the actual number of samples returned might be
        smaller than `samples` due to infinite or nan values.
    ret_idx: bool
        Also return a boolean array that corresponds to the
        downsampled indices in `a` and `b`.

    Returns
    -------
    dsa, dsb: 1d ndarrays of shape (samples,)
        The arrays `a` and `b` downsampled by evenly selecting
        points and pseudo-randomly adding or removing points
        to match `samples`.
    idx: 1d boolean array with same shape as `a`
        Only returned if `ret_idx` is True.
        A boolean array such that `a[idx] == dsa`
    """
    cdef uint32 samples_int = np.uint32(samples)

    # fixed random state for this method
    rs = np.random.RandomState(seed=47).get_state()

    keep = np.ones_like(a, dtype=bool)
    bad = np.isnan(a) | np.isinf(a) | np.isnan(b) | np.isinf(b)
    good = ~bad

    # We are generally not interested in bad data.
    keep[bad] = False
    bd = b[good]
    ad = a[good]

    cdef int64 diff

    if samples_int and samples_int < ad.size:
        # 1. Produce evenly distributed samples
        # Choosing grid-size:
        # - large numbers tend to show actual structures of the sample,
        #   which is not desired for plotting
        # - small numbers tend will not result in too few samples and,
        #   in order to reach the desired samples, the data must be
        #   upsampled again.
        # 300 is about the size of the plot in marker sizes and yields
        # good results.
        grid_size = 300
        # The events on the grid to process
        toproc = np.ones((grid_size, grid_size), dtype=np.uint8)

        x_discrete = np.array(norm(ad) * (grid_size - 1), dtype=np.uint32)
        y_discrete = np.array(norm(bd) * (grid_size - 1), dtype=np.uint32)

        # The events to keep
        keepd = np.zeros_like(ad, dtype=np.uint8)
        populate_grid(x_discrete=x_discrete,
                      y_discrete=y_discrete,
                      toproc=toproc,
                      keepd=keepd
                      )

        keepdb = np.array(keepd, dtype=bool)

        # 2. Make sure that we reach `samples` by adding or
        # removing events.
        diff = np.sum(keepdb) - samples_int
        if diff > 0:
            # Too many samples
            rem_indices = np.where(keepdb)[0]
            np.random.set_state(rs)
            rem = np.random.choice(rem_indices,
                                   size=diff,
                                   replace=False)
            keepdb[rem] = False
        elif diff < 0:
            # Not enough samples
            add_indices = np.where(~keepdb)[0]
            np.random.set_state(rs)
            add = np.random.choice(add_indices,
                                   size=abs(diff),
                                   replace=False)
            keepdb[add] = True

        # paulmueller 2024-01-03
        # assert np.sum(keepdb) <= samples_int, "sanity check"

        keep[good] = keepdb

        # paulmueller 2024-01-03
        # assert np.sum(keep) <= samples_int, "sanity check"

    if not remove_invalid:
        diff_bad = (samples_int or keep.size) - np.sum(keep)
        if diff_bad > 0:
            # Add a few of the invalid values so that in the end
            # we have the desired array size.
            add_indices_bad = np.where(bad)[0]
            np.random.set_state(rs)
            add_bad = np.random.choice(add_indices_bad,
                                       size=diff_bad,
                                       replace=False)
            keep[add_bad] = True

    # paulmueller 2024-01-03
    # if samples_int and not remove_invalid:
    #     assert np.sum(keep) == samples_int, "sanity check"

    asd = a[keep]
    bsd = b[keep]

    if ret_idx:
        return asd, bsd, keep
    else:
        return asd, bsd


def populate_grid(x_discrete, y_discrete, keepd, toproc):
    # Py_ssize_t is the proper C type for Python array indices.
    cdef int iter_size = x_discrete.size
    cdef Py_ssize_t ii
    cdef uint32[:] x_view = x_discrete
    cdef uint32[:] y_view = y_discrete
    # Numpy uses uint8 internally to represent boolean arrays
    cdef uint8[:] keepd_view = keepd
    cdef uint8[:, :] toproc_view = toproc

    for ii in range(iter_size):
        # filter for overlapping events
        xi = x_view[ii]
        yi = y_view[ii]
        if toproc_view[xi, yi]:
            toproc_view[xi, yi] = 0
            # include event
            keepd_view[ii] = 1


def valid(a, b):
    """Check whether `a` and `b` are not inf or nan"""
    return ~(np.isnan(a) | np.isinf(a) | np.isnan(b) | np.isinf(b))


def norm(a):
    """Normalize `a` with its min/max values"""
    rmin = a.min()
    rptp = a.max() - rmin
    return (a - rmin) / rptp



"""RT-DC hdf5 format"""
from __future__ import annotations

import warnings

import numbers
import numpy as np

from ... import definitions as dfn
from ...util import copy_if_needed

from . import feat_defect


class H5ContourEvent:
    def __init__(self, h5group, length=None):
        self._length = length
        self.h5group = h5group
        # for hashing in util.obj2bytes
        self.identifier = (h5group.file.filename, h5group["0"].name)

    def __getitem__(self, key):
        if not isinstance(key, numbers.Integral):
            # slicing!
            indices = np.arange(len(self))[key]
            output = []
            # populate the output list
            for evid in indices:
                output.append(self.h5group[str(evid)][:])
            return output
        elif key < 0:
            return self.__getitem__(key + len(self))
        else:
            return self.h5group[str(key)][:]

    def __iter__(self):
        for idx in range(len(self)):
            yield self[idx]

    def __len__(self):
        if self._length is None:
            # computing the length of an H5Group is slow
            self._length = len(self.h5group)
        return self._length

    @property
    def dtype(self):
        return self.h5group["0"].dtype

    @property
    def shape(self):
        return len(self), np.nan, 2


class H5Events:
    def __init__(self, h5):
        self.h5file = h5
        # According to https://github.com/h5py/h5py/issues/1960, we always
        # have to keep a reference to the HDF5 dataset, otherwise it will
        # be garbage-collected immediately. In addition to caching the HDF5
        # datasets, we cache the wrapping classes in the `self._cached_events`
        # dictionary.
        self._cached_events = {}
        self._defective_features = {}
        self._features_list = None

    @property
    def _features(self):
        if self._features_list is None:
            self._features_list = sorted(self.h5file["events"].keys())
            # make sure that "trace" is not empty
            if ("trace" in self._features
                    and len(self.h5file["events"]["trace"]) == 0):
                self._features_list.remove("trace")
        return self._features_list

    def __contains__(self, key):
        return key in self.keys()

    def __getitem__(self, key):
        if key not in self._cached_events:
            # user-level checking is done in core.py
            assert dfn.feature_exists(key), f"Feature '{key}' does not exist!"
            data = self.h5file["events"][key]
            if key == "contour":
                length = self.h5file.attrs.get("experiment:event count")
                fdata = H5ContourEvent(data, length=length)
            elif key == "mask":
                fdata = H5MaskEvent(data)
            elif key == "trace":
                fdata = H5TraceEvent(data)
            elif data.ndim == 1:
                fdata = H5ScalarEvent(data)
            else:
                # for features like "image", "image_bg" and other non-scalar
                # ancillary features
                fdata = data
            self._cached_events[key] = fdata
        return self._cached_events[key]

    def __iter__(self):
        # dict-like behavior
        for key in self.keys():
            yield key

    def _is_defective_feature(self, feat):
        """Whether the stored feature is defective"""
        if feat not in self._defective_features:
            defective = False
            if (feat in feat_defect.DEFECTIVE_FEATURES
                    and feat in self._features):
                # feature exists in the HDF5 file
                # workaround machinery for sorting out defective features
                defective = feat_defect.DEFECTIVE_FEATURES[feat](self.h5file)
            self._defective_features[feat] = defective
        return self._defective_features[feat]

    def keys(self):
        """Returns list of valid features

        Checks for
        - defective features: whether the data in the HDF5 file is invalid
        - existing feature names: dynamic, depending on e.g. plugin features
        """
        features = []
        for key in self._features:
            # check for defective features
            if dfn.feature_exists(key) and not self._is_defective_feature(key):
                features.append(key)
        return features


class H5MaskEvent:
    """Cast uint8 masks to boolean"""

    def __init__(self, h5dataset):
        self.h5dataset = h5dataset
        # identifier required because "mask" is used for computation
        # of ancillary feature "contour".
        self.identifier = (self.h5dataset.file.filename, self.h5dataset.name)
        self.dtype = np.dtype(bool)

    def __array__(self, dtype=np.bool_, copy=copy_if_needed, *args, **kwargs):
        if dtype is not np.uint8:
            warnings.warn("Please avoid calling the `__array__` method of the "
                          "`H5MaskEvent`. It may consume a lot of memory.",
                          UserWarning)
        # One of the reasons why we implement __array__ is such that
        # the data exporter knows this object is sliceable
        # (see yield_filtered_array_stacks).
        return self.h5dataset.__array__(dtype=dtype, *args, **kwargs)

    def __getitem__(self, idx):
        return np.asarray(self.h5dataset[idx], dtype=bool)

    def __iter__(self):
        for idx in range(len(self)):
            yield self[idx]

    def __len__(self):
        return len(self.h5dataset)

    @property
    def attrs(self):
        return self.h5dataset.attrs

    @property
    def shape(self):
        return self.h5dataset.shape


class H5ScalarEvent(np.lib.mixins.NDArrayOperatorsMixin):
    def __init__(self, h5ds):
        """Lazy access to a scalar feature with cache"""
        self.h5ds = h5ds
        # for hashing in util.obj2bytes
        self.identifier = (self.h5ds.file.filename, self.h5ds.name)
        self._array = None
        self.ndim = 1  # matplotlib might expect this from an array
        # attrs
        self._ufunc_attrs = dict(self.h5ds.attrs)

    def __array__(self, dtype=None, copy=copy_if_needed, *args, **kwargs):
        if self._array is None:
            self._array = np.asarray(self.h5ds, *args, **kwargs)
        return np.array(self._array, dtype=dtype, copy=copy)

    def __getitem__(self, idx):
        return self.__array__()[idx]

    def __len__(self):
        return len(self.h5ds)

    def _fetch_ufunc_attr(self, uname, ufunc):
        """A wrapper for calling functions on the scalar feature data

        The ideas are:

        1. If there is a ufunc (max/mean/min) value stored in the dataset
           attributes, then use this one.
        2. If the ufunc is computed, it is cached permanently in
           self._ufunc_attrs
        """
        val = self._ufunc_attrs.get(uname, None)
        if val is None:
            val = ufunc(self.__array__())
            self._ufunc_attrs[uname] = val
        return val

    def max(self, *args, **kwargs):
        return self._fetch_ufunc_attr("max", np.nanmax)

    def mean(self, *args, **kwargs):
        return self._fetch_ufunc_attr("mean", np.nanmean)

    def min(self, *args, **kwargs):
        return self._fetch_ufunc_attr("min", np.nanmin)

    @property
    def dtype(self):
        return self.h5ds.dtype

    @property
    def shape(self):
        return self.h5ds.shape


class H5TraceEvent:
    def __init__(self, h5group):
        self.h5group = h5group
        self._num_traces = None
        self._shape = None

    def __getitem__(self, idx):
        return self.h5group[idx]

    def __contains__(self, item):
        return item in self.h5group

    def __len__(self):
        if self._num_traces is None:
            self._num_traces = len(self.h5group)
        return self._num_traces

    def __iter__(self):
        for key in sorted(self.h5group.keys()):
            yield key

    def keys(self):
        return self.h5group.keys()

    @property
    def shape(self):
        if self._shape is None:
            atrace = list(self.h5group.keys())[0]
            self._shape = tuple([len(self)] + list(self.h5group[atrace].shape))
        return self._shape



"""Test command-line interface dclab-condense"""
import dclab
from dclab import cli, new_dataset, rtdc_dataset, util, RTDCWriter

import h5py
import numpy as np
import pytest

from helper_methods import DCOR_AVAILABLE, retrieve_data


def test_check_suffix_disabled_condense():
    path_in_o = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path_in = path_in_o.with_suffix("")
    path_in_o.rename(path_in)
    assert path_in.suffix == ""
    with pytest.raises(ValueError, match="Unsupported file type"):
        cli.condense(path_in=path_in,
                     path_out=path_in.with_name("condensed.rtdc"))
    # but this should work:
    cli.condense(path_in=path_in,
                 path_out=path_in.with_name("condensed2.rtdc"),
                 check_suffix=False)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_condense():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("condensed.rtdc")

    ret = cli.condense(path_in=path_in, path_out=path_out)
    assert ret is None, "by default, this method should return 0 (exit 0)"
    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert "dclab-condense" in dsj.logs
        assert len(dsj)
        assert len(dsj) == len(ds0)
        for feat in dsj.features:
            assert np.all(dsj[feat] == ds0[feat])


def test_condense_basins_include_by_default_data_from_basins():
    """
    When condensing a dataset, data from the basin should be
    written to the output file by default.
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    h5path_out = h5path.with_name("condensed.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features=["deform", "area_um"])
        hw.store_basin(basin_name="example basin",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[h5path],
                       basin_descr="an example test basin",
                       )
        assert "aspect" in src["events"], "sanity check"

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "deform" in ds.features_innate
        assert "aspect" not in ds.features_innate
        assert "aspect" in ds.features_basin

    # condense the basin-based dataset
    cli.condense(path_in=h5path_small, path_out=h5path_out)

    with h5py.File(h5path_out) as h5:
        assert "deform" in h5["events"], "sanity check"
        assert "aspect" in h5["events"], "basin feature was copied"

    # The basin information should still be available
    with new_dataset(h5path_out) as ds:
        assert "image" in ds
        assert "image" in ds.features_basin
        assert "image" not in ds.features_innate
        assert "deform" in ds.features_innate, "sanity check"


def test_condense_basins_include_no_data_from_basins():
    """
    When condensing a dataset, data from the basin should not be
    written to the output file when `store_basin_features` is
    set to False.
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    h5path_out = h5path.with_name("condensed.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features=["deform", "area_um"])
        hw.store_basin(basin_name="example basin",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[h5path],
                       basin_descr="an example test basin",
                       )
        assert "aspect" in src["events"], "sanity check"

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "deform" in ds.features_innate
        assert "aspect" not in ds.features_innate
        assert "aspect" in ds.features_basin

    # condense the basin-based dataset
    cli.condense(path_in=h5path_small, path_out=h5path_out,
                 store_ancillary_features=True,
                 store_basin_features=False)

    with h5py.File(h5path_out) as h5:
        assert "deform" in h5["events"], "sanity check"
        assert "aspect" not in h5["events"], "basin feature was copied"

    # The basin information should still be available
    with new_dataset(h5path_out) as ds:
        assert "aspect" in ds
        assert "aspect" in ds.features_basin
        assert "aspect" not in ds.features_innate
        assert "deform" in ds.features_innate, "sanity check"


@pytest.mark.parametrize("store_basins", [True, False])
def test_condense_basins_internal(store_basins):
    """
    Internal basins should just be copied to the new file.
    However, image data are not scalar data, so they should not
    show up in the condensed file.
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("input_file.rtdc")
    h5path_out = h5path.with_name("condensed.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        assert "basins" not in hw.h5file, "no basins in input file"
        # store scalar and non-scalar internal basins in the input file
        hw.store_basin(basin_name="scalar and non-scalar basin data",
                       basin_type="internal",
                       basin_format="h5dataset",
                       basin_locs=["basin_events"],
                       basin_descr="an example test basin",
                       internal_data={"userdef1": np.arange(2),
                                      "image_bg": np.zeros((2, 80, 320)),
                                      },
                       basin_map=np.zeros(src["events/deform"].shape[0]),
                       basin_feats=["image_bg", "userdef1"],
                       )

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "userdef1" in ds.features_basin
        assert "userdef1" not in ds.features_innate
        assert "image_bg" in ds.features_basin
        assert "image_bg" not in ds.features_innate

    # condense the basin-based dataset
    cli.condense(path_in=h5path_small,
                 path_out=h5path_out,
                 store_basin_features=store_basins)

    with h5py.File(h5path_out) as h5:
        assert "deform" in h5["events"], "sanity check"
        # The userdef1 feature should, in any case, not be in "events",
        # because the condense step always copies basin information,
        # including any internal basins.
        assert "userdef1" not in h5["events"]
        assert "userdef1" in h5["basin_events"]
        assert np.all(h5["basin_events"]["userdef1"] == np.arange(2))
        # The image_bg feature is not scalar, so it should not be here
        assert "image_bg" not in h5["events"]
        assert "image_bg" not in h5["basin_events"]

    with new_dataset(h5path_out) as ds:
        # userdef1 is an internal, scalar basin
        assert "userdef1" in ds.features_basin
        # image_bg is not scalar, so it should not be here
        assert "image_bg" not in ds.features_basin


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_condense_defective_feature():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")

    # The input file has an invalid volume
    with h5py.File(path_in, "a") as h5:
        bad_volume = h5["events/volume"][:]

    # compute correct volume with dclab
    with dclab.new_dataset(path_in) as ds:
        volume = ds["volume"][:]

    # dclab computes the correct volume, identifying the volume in the
    # file as defective.
    assert not np.allclose(volume, bad_volume, atol=0, rtol=1e-3), "sanity"

    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("condensed.rtdc")

    cli.condense(path_in=path_in, path_out=path_out)

    with h5py.File(path_out) as h5o:
        volume_out = h5o["events/volume"][:]

    # The output volume should be identical to the correct volume
    assert np.allclose(volume_out, volume, atol=0, rtol=1e-10)

    # Once again with dclab
    with dclab.new_dataset(path_out) as dso:
        assert np.allclose(dso["volume"], volume, atol=0, rtol=1e-10)


@pytest.mark.skipif(not DCOR_AVAILABLE, reason="DCOR is not available")
def test_condense_from_s3(tmp_path):
    """
    dclab 0.57.6 supports condensing any class of dataset. Here
    we just test whether we can condense a resource on DCOR
    """
    pytest.importorskip("requests")
    # TODO: Upload a smaller test dataset to DCOR to speed-up this test
    s3_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
              "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
              "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")

    path_cond = tmp_path / "condensed.rtdc"
    with new_dataset(s3_url) as ds, h5py.File(path_cond, "w") as h5_cond:
        cli.condense_dataset(ds=ds,
                             h5_cond=h5_cond,
                             store_ancillary_features=True)

    with new_dataset(path_cond) as dsc:
        assert "volume" in dsc
        assert np.allclose(dsc["deform"][1000], 0.0148279015,
                           atol=0, rtol=1e-7)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_condense_include_ancillary_features():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("condensed.rtdc")

    with h5py.File(path_in, "a") as h5:
        del h5["events/area_um"]

    cli.condense(path_in=path_in,
                 path_out=path_out,
                 store_ancillary_features=False)
    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert "volume" in ds0.features
        assert "volume" not in ds0.features_innate
        assert "contour" not in dsj.features
        assert "image" not in dsj.features
        assert "volume" not in dsj.features, "expensive feature not computed"
        assert "area_um" not in ds0.features_innate, "sanity check"
        assert "area_um" in ds0.features
        assert "area_um" in dsj.features_innate, "cheap feature available"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_condense_no_ancillary_features_control():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("condensed.rtdc")

    with h5py.File(path_in, "a") as h5:
        del h5["events/area_um"]

    cli.condense(path_in=path_in, path_out=path_out)  # defaults to True
    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert "volume" in ds0.features
        assert "volume" not in ds0.features_innate
        assert "volume" in dsj.features
        assert "area_um" not in ds0.features_innate, "sanity check"
        assert "area_um" in ds0.features
        assert "area_um" in dsj.features_innate


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_condense_triple_logs():
    """In version 0.57.6, we introduced log renaming"""
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("condensed.rtdc")
    path_out_2 = path_in.with_name("condensed2.rtdc")
    path_out_3 = path_in.with_name("condensed3.rtdc")

    cli.condense(path_in=path_in, path_out=path_out)
    cli.condense(path_out=path_out_2, path_in=path_out)
    cli.condense(path_out=path_out_3, path_in=path_out_2)

    with dclab.new_dataset(path_out) as ds, \
        dclab.new_dataset(path_out_2) as ds2, \
            dclab.new_dataset(path_out_3) as ds3:
        md5_meta_1 = util.hashobj(ds.config)
        assert f"dclab-condense_{md5_meta_1}" in list(ds2.logs)
        assert f"dclab-condense_{md5_meta_1}" in list(ds3.logs)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_condense_wo_logs():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    with h5py.File(path_in, "a") as h5:
        del h5["logs"]
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("condensed.rtdc")

    cli.condense(path_in=path_in, path_out=path_out)
    with new_dataset(path_out) as ds:
        assert len(ds.logs) == 2
        assert "dclab-condense" in ds.logs
        assert "dclab-condense-warnings" in ds.logs



"""Hooks for the HE-3D-FEM-22 dataset

https://doi.org/10.6084/m9.figshare.20630940.v1

See the individual functions for details.
"""
import matplotlib.pylab as plt
import numpy as np
from scipy import ndimage
from skimage import morphology

from .common import find_upper_intersection


def isoelastics_postprocess(lup, isoel):
    """Process extracted isoelastics"""
    return isoel


def lut_preprocess(lup):
    """Extrapolate isoelasticity lines to convex hull

    The extrapolation has similar results as shown in
    https://github.com/ZELLMECHANIK-DRESDEN/dclab/issues/191.
    """
    xmax = lup.lut_raw[:, 0].max()
    num_isoel = 33

    # Perform linear interpolation until convex hull.
    # don't use small deformation values for isoelastics computation

    lut_iso = lup.lut_raw[
        np.logical_or(lup.lut_raw[:, 1] > 0.03,
                      lup.lut_raw[:, 0] > 120)
    ]

    emod, mask_sim = lup.map_lut_to_grid(lut_iso)
    # Find points that should not be in that 2D `emod` array.
    # (bad interpolation from convex vs raw enclosing polygon)
    # Apply a closing disk filter
    # Zero-pad the mask beforehand (otherwise the disk filter is not
    # properly applied to pixels at the boundary of the image)
    ds = 20  # disk closing size
    mask_padded = np.pad(mask_sim, ((ds, ds), (ds, ds)))
    mask_padded_disk = morphology.binary_closing(
        mask_padded, footprint=morphology.disk(ds))
    # Fill any holes (in case of sparse simulations)
    ndimage.binary_fill_holes(mask_padded_disk, output=mask_padded_disk)
    mask_disk = mask_padded_disk[ds:-ds, ds:-ds]

    emod[~mask_disk] = np.nan
    levels, contours, contours_px = lup.extract_isoelastics_from_grid(
        emod, lut=lut_iso, num=num_isoel)

    # Extend the LUT via extrapolation
    lut_new = lup.lut_raw
    ncs = []
    if lup.verbose:
        ax1 = plt.subplot(121, title="extrapolation region")
        ax1.imshow(1.*mask_disk + mask_sim)
        ax2 = plt.subplot(122, title="LUT extrapolation")
        plt.plot(lup.lut_raw[:, 0], lup.lut_raw[:, 1], ".")
    for lev, cc in zip(levels, contours):
        if np.isnan(lev) or len(cc) < 10:
            continue
        ip = np.polynomial.Polynomial.fit(cc[:, 0], cc[:, 1], 3)
        new_cc = np.zeros((20, 3))
        xrange = (cc[-1, 0], 1.2*cc[-1, 0])
        new_cc[:, 0] = np.linspace(xrange[0], xrange[1], 20)
        new_cc[:, 1] = ip(new_cc[:, 0])
        new_cc[:, 2] = lev
        inside = lup.points_in_convex_hull(new_cc[:, :2])
        # Add a few points outside the convex hull for isoelastics computation
        for ii in range(len(inside)):
            if not inside[ii]:
                inside[ii:ii+4] = True
                break
        parts = new_cc[inside]
        lut_new = np.concatenate((lut_new, parts))
        # Append boundary point
        sx, sy = find_upper_intersection(
            ip, lup.convex_hull, xrange=xrange)
        new_cc = np.concatenate((new_cc, [[sx, sy, lev]]))
        ncs.append(new_cc)

        if lup.verbose:
            xtest = np.linspace(cc[0, 0], 1.5*cc[-1, 0], 20, endpoint=True)
            ax2.plot(cc[:, 0], cc[:, 1], "o")
            ax2.plot(xtest, ip(xtest))
            ax2.plot(new_cc[:, 0], new_cc[:, 1], "x")

    if lup.verbose:
        ax2.plot(lup.convex_hull[:, 0], lup.convex_hull[:, 1])
        ax2.set_ylim(0, 0.2)
        ax2.set_xlim(0, xmax)
        plt.show()
    lup.lut_raw = lut_new



import warnings

import numpy as np

import dclab
from dclab.rtdc_dataset import new_dataset

import pytest

from helper_methods import example_data_dict


def test_changed_polygon_filter():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    amin, amax = ds["area_um"].min(), ds["area_um"].max()
    dmin, dmax = ds["deform"].min(), ds["deform"].max()
    pf = dclab.PolygonFilter(axes=["area_um", "deform"],
                             points=[[amin, dmin],
                                     [(amax + amin) / 2, dmin],
                                     [(amax + amin) / 2, dmax],
                                     ])
    ds.config["filtering"]["polygon filters"].append(pf.unique_id)
    ds.apply_filter()
    assert np.sum(ds.filter.all) == 2138
    # change the filter
    pf.points = list(pf.points) + [np.array([amin, dmax])]
    ds.apply_filter()
    assert np.sum(ds.filter.all) == 4215
    # invert the filter
    pf.inverted = True
    ds.apply_filter()
    assert np.sum(ds.filter.all) == 4257


def test_disable_filters():
    """Disabling the filters should only affect RTDCBase.filter.all"""
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.filter.manual[[0, 8471]] = False
    ds.apply_filter()
    ds.config["filtering"]["enable filters"] = False
    ds.apply_filter()
    assert np.all(ds.filter.all)


def test_filter_manual():
    # make sure min/max values are filtered
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.filter.manual[[0, 8471]] = False
    ds.apply_filter()
    assert len(ds["deform"][ds.filter.all]) == 8470
    assert ds["deform"][1] == ds["deform"][ds.filter.all][0]


def test_filter_min_max():
    # make sure min/max values are filtered
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    amin, amax = ds["area_um"].min(), ds["area_um"].max()
    ds.config["filtering"]["area_um min"] = (amax + amin) / 2
    ds.config["filtering"]["area_um max"] = amax
    ds.apply_filter()
    assert np.sum(ds.filter.all) == 4256

    # make sure data is not filtered before calling ds.apply_filter
    dmin, dmax = ds["deform"].min(), ds["deform"].max()
    ds.config["filtering"]["deform min"] = (dmin + dmax) / 2
    ds.config["filtering"]["deform max"] = dmax
    assert np.sum(ds.filter.all) == 4256


def test_nan_warning():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ddict["area_um"][[1, 4, 6]] = np.nan
    ds = new_dataset(ddict)
    amin, amax = np.nanmin(ds["area_um"]), np.nanmax(ds["area_um"])
    ds.config["filtering"]["area_um min"] = (amax + amin) / 2
    ds.config["filtering"]["area_um max"] = amax
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        ds.apply_filter()
        for wi in w:  # sometimes there are ResourceWarnings
            if wi.category == dclab.rtdc_dataset.filter.NanWarning:
                assert str(wi.message).count("area_um")
                break
        else:
            assert False, "Expected NanWarning"
    assert np.sum(ds.filter.all) == 4255


def test_only_one_boundary_error():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ddict["area_um"][[1, 4, 6]] = np.nan
    ds = new_dataset(ddict)
    amin, amax = np.nanmin(ds["area_um"]), np.nanmax(ds["area_um"])
    ds.config["filtering"]["area_um min"] = (amax + amin) / 2
    try:
        ds.apply_filter()
    except ValueError:
        pass
    else:
        assert False, "setting only half of a box filter should not work"


@pytest.mark.parametrize("prop", ["all", "box", "invalid", "polygon"])
def test_readonly_properties(prop):
    ddict = example_data_dict(size=8472, keys=["area_um", "deform",
                                               "emodulus"])
    ds = new_dataset(ddict)
    thing = getattr(ds.filter, prop)
    with pytest.raises(ValueError,
                       match="assignment destination is read-only"):
        thing[:10] = False


def test_remove_ancillary_feature():
    """When a feature is removed, the box boolean filter must be deleted"""
    ddict = example_data_dict(size=8472, keys=["area_um", "deform",
                                               "emodulus"])
    ds = new_dataset(ddict)
    amin, amax = ds["area_um"].min(), ds["area_um"].max()
    ds.config["filtering"]["area_um min"] = (amax + amin) / 2
    ds.config["filtering"]["area_um max"] = amax
    emin, emax = ds["emodulus"].min(), ds["emodulus"].max()
    ds.config["filtering"]["emodulus min"] = (emax + emin) / 2
    ds.config["filtering"]["emodulus max"] = emax
    ds.apply_filter()
    numevents = np.sum(ds.filter.all)
    # now remove the ancillary feature
    ds._events.pop("emodulus")
    ds.apply_filter()
    numevents2 = np.sum(ds.filter.all)
    assert numevents != numevents2



import json
import time

from ...http_utils import REQUESTS_AVAILABLE  # noqa: F401
from ...http_utils import requests, session_cache


class DCORAccessError(BaseException):
    pass


class APIHandler:
    """Handles the DCOR api with caching for simple queries"""
    #: these are cached to minimize network usage
    cache_queries = ["metadata", "size", "feature_list", "valid"]
    #: DCOR API Keys/Tokens in the current session
    api_keys = []

    def __init__(self, url, api_key="", cert_path=None, dcserv_api_version=2):
        """

        Parameters
        ----------
        url: str
            URL to DCOR API
        api_key: str
            DCOR API token
        cert_path: pathlib.Path
            the path to the server's CA bundle; by default this
            will use the default certificates (which depends on
            from where you obtained certifi/requests)
        """
        #: DCOR API URL
        self.url = url
        #: keyword argument to :func:`requests.request`
        self.verify = cert_path or True
        #: DCOR API Token
        self.api_key = api_key
        #: ckanext-dc_serve dcserv API version
        self.dcserv_api_version = dcserv_api_version
        #: create a session
        self.session = session_cache.get_session(url)
        self._cache = {}

    @classmethod
    def add_api_key(cls, api_key):
        """Add an API Key/Token to the base class

        When accessing the DCOR API, all available API Keys/Tokens are
        used to access a resource (trial and error).
        """
        if api_key.strip() and api_key not in APIHandler.api_keys:
            APIHandler.api_keys.append(api_key)

    def _get(self, query, feat=None, trace=None, event=None, api_key="",
             retries=13):
        # "version=2" introduced in dclab 0.54.3
        # (supported since ckanext.dc_serve 0.13.2)
        qstr = f"&version={self.dcserv_api_version}&query={query}"
        if feat is not None:
            qstr += f"&feature={feat}"
        if trace is not None:
            qstr += f"&trace={trace}"
        if event is not None:
            qstr += f"&event={event}"
        apicall = self.url + qstr
        fail_reasons = []
        for _ in range(retries):
            try:
                # try-except both requests and json conversion
                req = self.session.get(apicall,
                                       headers={"Authorization": api_key},
                                       verify=self.verify,
                                       timeout=1,
                                       )
                jreq = req.json()
            except requests.urllib3.exceptions.ConnectionError:  # requests
                fail_reasons.append("connection problem")
                continue
            except (requests.urllib3.exceptions.ReadTimeoutError,
                    requests.exceptions.ConnectTimeout):  # requests
                fail_reasons.append("timeout")
            except json.decoder.JSONDecodeError:  # json
                fail_reasons.append("invalid json")
                time.sleep(1)  # wait a bit, maybe the server is overloaded
                continue
            else:
                break
        else:
            raise DCORAccessError(f"Could not complete query '{apicall}'. "
                                  f"I retried {retries} times. "
                                  f"Messages: {fail_reasons}")
        return jreq

    def get(self, query, feat=None, trace=None, event=None):
        if query in APIHandler.cache_queries and query in self._cache:
            result = self._cache[query]
        else:
            req = {"error": {"message": "No access to API (api key?)"}}
            for api_key in [self.api_key] + APIHandler.api_keys:
                req = self._get(query, feat, trace, event, api_key)
                if req["success"]:
                    self.api_key = api_key  # remember working key
                    break
            else:
                raise DCORAccessError(
                    f"Cannot access {query}: {req['error']['message']}")
            result = req["result"]
            if query in APIHandler.cache_queries:
                self._cache[query] = result
        return result




import numpy as np

from dclab import downsampling


def test_basic():
    a = np.arange(100)
    b, idx = downsampling.downsample_rand(a=a,
                                          samples=5,
                                          ret_idx=True)
    assert np.all(a[idx] == b)

    b2 = downsampling.downsample_rand(a=a,
                                      samples=5,
                                      ret_idx=False)
    assert np.all(b2 == b)


def test_grid_basic():
    a = np.arange(100)
    b = np.arange(50, 150)
    ads, bds, idx = downsampling.downsample_grid(a=a,
                                                 b=b,
                                                 samples=5,
                                                 ret_idx=True)
    assert np.all(a[idx] == ads)
    assert np.all(b[idx] == bds)

    ads2, bds2 = downsampling.downsample_grid(a=a,
                                              b=b,
                                              samples=5,
                                              ret_idx=False)
    assert np.all(ads2 == ads)
    assert np.all(bds2 == bds)


def test_grid_nan():
    a = np.arange(100, dtype=float)
    a[50:] = np.nan
    b = np.arange(50, 150, dtype=float)
    ads, bds, idx = downsampling.downsample_grid(a=a,
                                                 b=b,
                                                 samples=52,
                                                 ret_idx=True,
                                                 remove_invalid=False)
    assert np.sum(idx) == 52
    assert np.allclose(a[idx], ads, atol=1e-14, rtol=0, equal_nan=True)
    assert np.allclose(b[idx], bds, atol=1e-14, rtol=0, equal_nan=True)
    assert np.sum(np.isnan(ads)) == 2, "2 more samples requested than non-nan"

    ads2, bds2 = downsampling.downsample_grid(a=a,
                                              b=b,
                                              samples=50,
                                              ret_idx=False,
                                              remove_invalid=True)
    assert np.sum(np.isnan(ads2)) == 0
    assert np.all(ads2 == a[:50])
    assert np.all(bds2 == b[:50])

    ads3, bds3, idx3 = downsampling.downsample_grid(a=a,
                                                    b=b,
                                                    samples=60,
                                                    ret_idx=True,
                                                    remove_invalid=True)
    assert np.sum(np.isnan(bds3)) == 0
    assert ads3.size == 50, "only 50 valid values"
    assert np.all(ads3 == a[idx3])
    assert np.all(bds3 == b[idx3])
    assert np.all(bds3 == b[:50])


def test_grid_nan_keep_all():
    a = np.arange(100, dtype=float)
    a[50:] = np.nan
    b = np.arange(50, 150, dtype=float)
    ads, bds, idx = downsampling.downsample_grid(a=a,
                                                 b=b,
                                                 samples=0,
                                                 ret_idx=True,
                                                 remove_invalid=False)
    assert np.allclose(a, ads, equal_nan=True)
    assert np.allclose(b, bds, equal_nan=True)


def test_grid_nan_keep_all_except_invalid():
    a = np.arange(100, dtype=float)
    a[50:] = np.nan
    b = np.arange(50, 150, dtype=float)
    ads, bds, idx = downsampling.downsample_grid(a=a,
                                                 b=b,
                                                 samples=0,
                                                 ret_idx=True,
                                                 remove_invalid=True)
    assert np.allclose(a[:50], ads, equal_nan=True)
    assert np.allclose(b[:50], bds, equal_nan=True)


def test_rand_nan():
    a = np.arange(100, dtype=float)
    a[50:] = np.nan
    b, idx = downsampling.downsample_rand(a=a,
                                          samples=5,
                                          ret_idx=True,
                                          remove_invalid=False)
    assert np.allclose(a[idx], b, atol=1e-14, rtol=0, equal_nan=True)
    assert np.sum(np.isnan(b)) == 4

    b2, idx2 = downsampling.downsample_rand(a=a,
                                            samples=5,
                                            ret_idx=True,
                                            remove_invalid=True)
    assert np.allclose(a[idx2], b2, atol=1e-14, rtol=0, equal_nan=True)
    assert np.sum(np.isnan(b2)) == 0


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numpy as np

from ...util import hashobj

from ..filter import Filter

from .mapper import map_indices_root2child, map_indices_child2root


class HierarchyFilterError(BaseException):
    """Used for unexpected filtering operations"""
    pass


class HierarchyFilter(Filter):
    def __init__(self, rtdc_ds):
        """A filtering class for RTDC_Hierarchy

        This subclass handles manual filters for hierarchy children.
        The general problem with hierarchy children is that their data
        changes when the hierarchy parent changes. As hierarchy
        children may also have hierarchy children, dealing with
        manual filters (`Filter.manual`) is not trivial. Here,
        the manual filters are translated into event indices of the
        root parent (the highest member of the hierarchy, which is
        `RTDC_Hierarchy.hparent` if there is only one child).
        This enables to keep track of the manually excluded events
        even if

        - the parent changes its filters,
        - the parent is a hierarchy child as well, or
        - the excluded event is filtered out in the parent.
        """
        self._man_root_ids = []
        super(HierarchyFilter, self).__init__(rtdc_ds)
        self._parent_rtdc_ds = None
        self._parent_hash = None
        self.update_parent(rtdc_ds.hparent)

    @property
    def parent_changed(self):
        return hashobj(self._parent_rtdc_ds.filter.all) != self._parent_hash

    def apply_manual_indices(self, rtdc_ds, manual_indices):
        """Write to `self.manual`

        Write `manual_indices` to the boolean array `self.manual`
        and also store the indices as `self._man_root_ids`.

        Notes
        -----
        If `self.parent_changed` is `True`, i.e. the parent applied
        a filter and the child did not yet hear about this, then
        `HierarchyFilterError` is raised. This is important, because
        the size of the current filter would not match the size of
        the filtered events of the parent and thus index-mapping
        would not work.
        """
        if self.parent_changed:
            msg = "Cannot apply filter, because parent changed: " \
                  + "dataset {}. ".format(rtdc_ds) \
                  + "Run `RTDC_Hierarchy.apply_filter()` first!"
            raise HierarchyFilterError(msg)
        else:
            self._man_root_ids = list(manual_indices)
            cidx = map_indices_root2child(child=rtdc_ds,
                                          root_indices=manual_indices)
            if len(cidx):
                self.manual[cidx] = False

    def reset(self):
        super(HierarchyFilter, self).reset()
        self._man_root_ids.clear()

    def retrieve_manual_indices(self, rtdc_ds):
        """Read from self.manual

        Read from the boolean array `self.manual`, index all
        occurences of `False` and find the corresponding indices
        in the root hierarchy parent, return those and store them
        in `self._man_root_ids` as well.

        Notes
        -----
        This method also retrieves hidden indices, i.e. events
        that are not part of the current hierarchy child but
        which have been manually excluded before and are now
        hidden because a hierarchy parent filtered it out.

        If `self.parent_changed` is `True`, i.e. the parent applied
        a filter and the child did not yet hear about this, then
        nothing is computed and `self._man_root_ids` as-is.  This
        is important, because the size of the current filter would
        not match the size of the filtered events of the parent and
        thus index-mapping would not work.
        """
        if self.parent_changed:
            # ignore
            pass
        elif np.all(self.manual):
            # Do not do anything and remember the events we manually
            # excluded in case the parent reinserts them.
            pass
        else:
            # indices from boolean array
            pbool = map_indices_child2root(
                child=rtdc_ds,
                child_indices=np.where(~self.manual)[0]).tolist()
            # retrieve all indices that are currently not visible
            # previous indices
            pold = self._man_root_ids
            # all indices previously selected either via
            # - self.manual or
            # - self.apply_manual_indices
            pall = sorted(list(set(pbool + pold)))
            # visible indices (only available child indices are returned)
            pvis_c = map_indices_root2child(child=rtdc_ds,
                                            root_indices=pall).tolist()
            # map visible child indices back to root indices
            pvis_p = map_indices_child2root(child=rtdc_ds,
                                            child_indices=pvis_c).tolist()
            # hidden indices
            phid = list(set(pall) - set(pvis_p))
            # Why not set `all_idx` to `pall`:
            # - pbool is considered to be correct
            # - pold contains hidden indices, but also might contain
            #   excess indices from before, i.e. if self.apply_manual_indices
            #   is called, self.manual is also updated. If however,
            #   self.manual is updated, self._man_root_ids are not updated.
            #   Thus, we trust pbool (self.manual) and only use pold
            #   (self._man_root_ids) to determine hidden indices.
            all_idx = list(set(pbool + phid))
            self._man_root_ids = sorted(all_idx)
        return self._man_root_ids

    def update_parent(self, parent_rtdc_ds):
        # hold reference to rtdc_ds parent
        # (not to its filter, because that is reinstantiated)
        self._parent_rtdc_ds = parent_rtdc_ds
        self._parent_hash = hashobj(self._parent_rtdc_ds.filter.all)



import hashlib

from ..http_utils import HTTPFile, REQUESTS_AVAILABLE, is_url_available
from ..http_utils import is_http_url  # noqa: F401

from .feat_basin import Basin
from .fmt_hdf5 import RTDC_HDF5


class RTDC_HTTP(RTDC_HDF5):
    def __init__(self,
                 url: str,
                 *args, **kwargs):
        """Access RT-DC measurements via HTTP

        This class allows you to open .rtdc files accessible via an
        HTTP URL, for instance files on an S3 object storage or
        figshare download links.

        This is essentially just a wrapper around :class:`.RTDC_HDF5`
        with :class:`.HTTPFile` passing a file object to h5py.

        Parameters
        ----------
        url: str
            Full URL to an HDF5 file
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`

        Attributes
        ----------
        path: str
            The URL to the object

        Notes
        -----
        Since this format still requires random access to the file online,
        i.e. not the entire file is downloaded, only parts of it, the
        web server must support range requests.
        """
        if not REQUESTS_AVAILABLE:
            raise ModuleNotFoundError(
                f"Package `requests` required for loading http data '{url}'!")

        self._fhttp = HTTPFile(url)
        if kwargs.get("identifier") is None:
            if self._fhttp.etag is not None:
                # Set the HTTP ETag as the identifier, it doesn't get
                # more unique than that!
                kwargs["identifier"] = self._fhttp.etag
            else:
                # Compute a hash of the first data chunk
                kwargs["identifier"] = hashlib.md5(
                    self._fhttp.get_cache_chunk(0)).hexdigest()

        # Initialize the HDF5 dataset
        super(RTDC_HTTP, self).__init__(
            h5path=self._fhttp,
            *args,
            **kwargs)
        # Override self.path with the actual HTTP URL
        self.path = url

    def close(self):
        super(RTDC_HTTP, self).close()
        self._fhttp.close()


class HTTPBasin(Basin):
    basin_format = "http"
    basin_type = "remote"

    def __init__(self, *args, **kwargs):
        self._available_verified = None
        super(HTTPBasin, self).__init__(*args, **kwargs)

    def _load_dataset(self, location, **kwargs):
        h5file = RTDC_HTTP(location, **kwargs)
        return h5file

    def is_available(self):
        """Check for `requests` and object availability

        Caching policy: Once this method returns True, it will always
        return True.
        """
        if self._available_verified is None:
            with self._av_check_lock:
                if not REQUESTS_AVAILABLE:
                    # don't even bother
                    self._available_verified = False
                else:
                    avail, reason = is_url_available(self.location,
                                                     ret_reason=True)
                    if reason in ["forbidden", "not found"]:
                        # we cannot access the URL in the near future
                        self._available_verified = False
                    elif avail:
                        self._available_verified = True
        return self._available_verified



import numpy as np
import pytest

from dclab.features import emodulus


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_simple_emod_he_2d_fem_22():
    x = np.linspace(0, 250, 100)
    y = np.linspace(0, 0.1, 100)
    x, y = np.meshgrid(x, y)

    emod = emodulus.get_emodulus(area_um=x,
                                 deform=y,
                                 medium="CellCarrier",
                                 channel_width=30,
                                 flow_rate=0.16,
                                 lut_data="HE-2D-FEM-22",
                                 px_um=0,  # without pixelation correction
                                 temperature=23,
                                 visc_model="herold-2017")

    assert np.allclose(emod[10, 50], 1.2520069112713232)
    assert np.allclose(emod[50, 50], 0.6442165714687615)
    assert np.allclose(emod[80, 50], 0.575373042660183)

    assert np.allclose(emod[10, 80], 1.638297273096505)
    assert np.allclose(emod[50, 80], 0.8191788744746755)
    assert np.allclose(emod[80, 80], 0.716695164528623)
